/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * AI Tank Sensors
 *
 * An implementation of tank sensors that allows using "normal" Brains
 * as clients with TTWP. Just for testing
 */

package ToyTank;

public class TTWPSensor implements TankSensor {

    static final int LONG_RANGE_DISTANCE = 512;
    static final int LONG_RANGE_START = 16;
    static final int LONG_RANGE_SEGMENTS = 6;
    static final int LONG_RANGE_SECTORS = 6;

    static final int SHORT_RANGE_DISTANCE = 16;
    static final int SHORT_RANGE_SECTORS = 12;

    static final int COLLISION_SECTORS = 6;

    static final char NONE = '0';
    static final char OBSTACLE = '1';
    static final char PHYSICAL = '2';
    static final char BOTH = '3';

    public String energy;
    public String aim;
    public String collision;
    public String shortRange;
    public String longRange;

    // Math
    private static final float PI = (float) Math.PI;
    private static final float TWOPI = (float) (2 * Math.PI);
    private static final float TO_RADIANS = PI / 180f;

    private static final float[] longRangeSectors = {
        -PI / 3, -PI / 6, -PI / 18, 0, PI / 18, PI / 6, PI / 3
    };

    private static final float[] longRangeSegments = {
        16, 32, 64, 128, 256, 512
    };

    private static float shortRangeSector = TWOPI / SHORT_RANGE_SECTORS;

    private static float collisionSector = TWOPI / COLLISION_SECTORS;

    static {

        // Use normalised angles everywhere
        for (int i = 0; i < longRangeSectors.length; i++)
            longRangeSectors[i] = normalise(longRangeSectors[i]);
    }

    TTWPSensor(String e, String a, String c, String s, String l) {
        energy = e;
        aim = a;
        collision = c;
        shortRange = s;
        longRange = l;
    }

    public int energy() {
        return (new Integer(energy)).intValue();
    }

    public boolean turretRadar(float start, float finish, float range) {

        start = normalise(start);
        finish = normalise(finish);

        if (lessEq(finish, start) || range <= 0)
            return false;

            // Scan image, sector by sector

        for (int x = 0; x < longRangeSectors.length - 1; x++) {

            // Is this sector included in the search angle?
            if (less(start, longRangeSectors[x + 1])
                    && less(longRangeSectors[x], finish)) {

                // Scan this sector
                for (int y = 0; y < longRangeSegments.length; y++) {
                    if (y == 0 || range >= longRangeSegments[y - 1]) {

                        if (physical(longRange.charAt(x
                                                      + (LONG_RANGE_SEGMENTS
                                                         * y))))
                            return true;
                    }
                    else {
                        break;
                    }
                }
            }
        }

        return false;
    }

    public boolean tankRadar(float start, float finish, float range) {

        start = normalise(start);
        finish = normalise(finish);

        if (lessEq(finish, start))
            return false;

            // Scan image, sector by sector

        for (int x = 0; x < SHORT_RANGE_SECTORS; x++) {

            // Is this sector included in the search angle?
            if (less(start, (x + 1) * shortRangeSector)
                    && less(x * shortRangeSector, finish)) {

                if (physical(shortRange.charAt(x)))
                    return true;
            }
        }

        return false;
    }

    public boolean obstacleSensor(float start, float finish, float range) {

        start = normalise(start);
        finish = normalise(finish);

        if (lessEq(finish, start))
            return false;

            // Scan image, sector by sector

        for (int x = 0; x < SHORT_RANGE_SECTORS; x++) {

            // Is this sector included in the search angle?
            if (less(start, (x + 1) * shortRangeSector)
                    && less(x * shortRangeSector, finish)) {

                if (obstacle(shortRange.charAt(x)))
                    return true;
            }
        }

        return false;
    }

    public boolean collisionSensor(float start, float finish) {

        start = normalise(start);
        finish = normalise(finish);

        if (lessEq(finish, start))
            return false;

            // Scan image, sector by sector

        for (int x = 0; x < COLLISION_SECTORS; x++) {

            // Is this sector included in the search angle?
            if (less(start, (x + 1) * collisionSector)
                    && less(x * collisionSector, finish)) {

                if (obstacle(collision.charAt(x)))
                    return true;
            }
        }

        return false;
    }


    /* Utility methods */

    private static boolean obstacle(char c) {
        return (c == OBSTACLE) || (c == BOTH);
    }

    private static boolean physical(char c) {
        return (c == PHYSICAL) || (c == BOTH);
    }

    // Is first angle less than second
    private static boolean less(float first, float second) {
        if (first < second)
            return first + PI > second;
        else
            return first > second + PI;
    }

    // Is first angle less or equal than second
    private static boolean lessEq(float first, float second) {
        if (first < second)
            return first + PI >= second;
        else
            return first >= second + PI;
    }

    // Ensure angle is between [0..TWOPI]
    private static float normalise(float angle) {
        if (angle < 0)
            return angle + TWOPI;
        else if (angle > TWOPI)
            return angle - TWOPI;
        else
            return angle;
    }

}

