/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * pmarco@ipvvis.unipv.it
 *
 * Message handling on client side is oversimplified.
 */

package ToyTank;

import java.io.*;
import java.awt.Color;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;


public final class TTWProtocol {

    // Debugging
    private static boolean debug = false;

    static {
        try {

            // Set to true if defined
            if (System.getProperty("ToyTankWorld.DEBUG", "FALSE") != "FALSE")
                debug = true;
        }
        catch (Exception e) {

            // Someone (e.g. a browser) may not like the above
        }
    }

    static void setDebug(boolean d) {
        debug = d;
    }

    // Commands
    public static final String BANNER =
        "ToyTankWorldProtocol (TTWP) - Vers. 1.1";
    public static final String DESCRIPTION = "DESCRIPTION";
    public static final String SENSOR = "SENSOR";
    public static final String ACTION = "ACTION";
    public static final String STOP = "STOP";
    public static final String CLOSE = "CLOSE";
    public static final String RESTART = "RESTART";

    // Acknowledgement
    public static final String OK = "200 OK";
    public static final String ERROR = "500 ERROR";

    // Communication
    private BufferedReader inStream;
    private PrintStream outStream;


    public TTWProtocol(BufferedReader in, PrintStream out) {
        inStream = in;
        outStream = out;
    }


    public void readDescription(TankDescription t) throws IOException {

        // Send banner message
        outStream.println(BANNER);
        outStream.flush();

        if (debug)
            System.err.println("TTWP.out: " + BANNER);

            // Receive description

        String inputLine = null;

        try {
            inputLine = inStream.readLine();

            if (debug)
                System.err.println("TTWP.in: " + inputLine);

            StringTokenizer input = new StringTokenizer(inputLine, " \n");

            try {
                String inputToken = input.nextToken();

                // Start parsing
                if (!inputToken.equals(DESCRIPTION))
                    throw new TTWPException();

                t.setName(input.nextToken());

                int r, g, b;

                try {
                    r = Integer.valueOf(input.nextToken()).intValue();
                    g = Integer.valueOf(input.nextToken()).intValue();
                    b = Integer.valueOf(input.nextToken()).intValue();
                }
                catch (NumberFormatException e) {
                    throw new TTWPException();
                }
                catch (RuntimeException e) {

                    // Color parameters could be out of range
                    throw new TTWPException();
                }

                t.setColor(new Color(r, g, b));
            }
            catch (NoSuchElementException e) {
                throw new TTWPException();
            }
        }
        catch (TTWPException e) {

            // Turn it into am IOException, but notify client first
            outStream.println(ERROR);
            outStream.flush();

            if (debug)
                System.err.println("TTWP.out: " + ERROR);

            throw new IOException("Illegal command: " + inputLine);
        }

        // Send ACK
        outStream.println(OK);
        outStream.flush();

        if (debug)
            System.err.println("TTWP.out: " + OK);
    }

    public void writeDescription(String name, int r, int g,
                                 int b) throws IOException {

        // Wait for an incoming message
        String inputLine = inStream.readLine();

        if (debug)
            System.err.println("TTWP.in: " + inputLine);

            // ... which is a banner message, not relevant for now.
            // Then send the description

        StringBuffer command = new StringBuffer(DESCRIPTION);

        command.append(" ").append(name);
        command.append(" ").append(r);
        command.append(" ").append(g);
        command.append(" ").append(b);
        outStream.println(command);
        outStream.flush();

        if (debug)
            System.err.println("TTWP.out: " + command);

            // Wait for confirmation

        inputLine = inStream.readLine();

        if (debug)
            System.err.println("TTWP.in: " + inputLine);

            // Validate

        if (!inputLine.startsWith(OK))
            throw new IOException(inputLine);
    }

    public String readCommand() throws IOException {

        // Receive sensor command
        String inputLine = null;

        inputLine = inStream.readLine();

        if (debug)
            System.err.println("TTWP.in: " + inputLine);

        return inputLine;
    }


    public TankSensor readSensor(String inputLine) throws IOException {

        // Receive sensor command
        try {
            StringTokenizer input = new StringTokenizer(inputLine, " \n");

            try {
                String inputToken = input.nextToken();

                // Start parsing
                if (!inputToken.equals(SENSOR))
                    throw new TTWPException();

                String energy = input.nextToken();
                String aim = input.nextToken();
                String collision = input.nextToken();
                String shortRange = input.nextToken();
                String longRange = input.nextToken();

                /*
                // For debugging: printout what the tank 'sees'
                if (debug) {
                    System.err.println("Turret view:");

                    for (int i = 5; i >= 0; i--) {
                        System.err.println(longRange.substring(6 * i,
                                                               6 * (i + 1)));
                    }
                }
                */

                TTWPSensor res = new TTWPSensor(energy, aim, collision,
                                                shortRange, longRange);

                return res;
            }
            catch (NoSuchElementException e) {
                throw new TTWPException();
            }
        }
        catch (TTWPException e) {

            // Turn it into an IOException, but notify client first
            outStream.println(ERROR);
            outStream.flush();

            if (debug)
                System.err.println("TTWP.out: " + ERROR);

            throw new IOException("Illegal command: " + inputLine);
        }

    }

    public void writeSensor(TTWPSensor t) {

        StringBuffer command = new StringBuffer(SENSOR);

        command.append(" ").append(t.energy);
        command.append(" ").append(t.aim);

        // Collision view
        command.append(" ");
        command.append(t.collision);

        // Short range view
        command.append(" ");
        command.append(t.shortRange);

        // Long range view
        command.append(" ");
        command.append(t.longRange);

        outStream.println(command);
        outStream.flush();

        if (debug)
            System.err.println("TTWP.out: " + command);

            // NO confirmation will be sent back
    }

    public void readAction(TankAction t) throws IOException {

        // Receive action dataset
        String inputLine = null;

        try {
            inputLine = inStream.readLine();

            if (debug)
                System.err.println("TTWP.in: " + inputLine);

            StringTokenizer input = new StringTokenizer(inputLine, " \n");

            try {
                String inputToken = input.nextToken();

                // Start parsing
                if (!inputToken.equals(ACTION))
                    throw new TTWPException();

                try {

                    // Read values
                    float turn =
                        Float.valueOf(input.nextToken()).floatValue();
                    float move =
                        Float.valueOf(input.nextToken()).floatValue();
                    float aim = Float.valueOf(input.nextToken()).floatValue();
                    String fire = input.nextToken();

                    // Transfer action; Limits are checked for in class Tank
                    t.turn(turn);
                    t.move(move);
                    t.aim(aim);

                    if (fire.equals("1"))
                        t.fire();
                    else if (fire.equals("0"))

                        /* do nothing */
                        ;
                    else
                        throw new TTWPException();
                }
                catch (NumberFormatException e) {
                    throw new TTWPException();
                }
            }
            catch (NoSuchElementException e) {
                throw new TTWPException();
            }
        }
        catch (TTWPException e) {

            // Turn it into an IOException, but notify client first
            outStream.println(ERROR);
            outStream.flush();

            if (debug)
                System.err.println("TTWP.out: " + ERROR);

            throw new IOException("Illegal command: " + inputLine);
        }
    }

    public void writeAction(Brain b) throws IOException {
        TTWPAction res = new TTWPAction();

        b.action(res);

        StringBuffer command = new StringBuffer(ACTION);

        command.append(" ").append(res.turn);
        command.append(" ").append(res.move);
        command.append(" ").append(res.aim);
        command.append(" ").append(res.fire ? "1" : "0");

        outStream.println(command);
        outStream.flush();

        if (debug)
            System.err.println("TTWP.out: " + command);

            // NO confirmation will be sent back
    }

    public void writeRestart() throws IOException {

        StringBuffer command = new StringBuffer(RESTART);

        outStream.println(command);
        outStream.flush();

        if (debug)
            System.err.println("TTWP.out: " + command);
    }

    public void writeStop() throws IOException {

        StringBuffer command = new StringBuffer(STOP);

        outStream.println(command);
        outStream.flush();

        if (debug)
            System.err.println("TTWP.out: " + command);
    }

    public void writeClose(int finalGap) throws IOException {

        StringBuffer command = new StringBuffer(CLOSE);

        command.append(" ").append(finalGap);

        outStream.println(command);
        outStream.flush();

        if (debug)
            System.err.println("TTWP.out: " + command);
    }

    public int readFinalGap(String inputLine) throws IOException {
        int finalGap = 0;

        try {
            StringTokenizer input = new StringTokenizer(inputLine, " \n");

            String inputToken = input.nextToken();

            // Start parsing
            if (!inputToken.equals(CLOSE))
                throw new TTWPException();

            finalGap = Integer.valueOf(input.nextToken()).intValue();
        }
        catch (NumberFormatException e) {
            throw new IOException("Illegal command: " + inputLine);
        }
        catch (TTWPException e) {
            throw new IOException("Illegal command: " + inputLine);
        }

        return finalGap;
    }

    public boolean isSensorCommand(String inputLine) {
        return inputLine.startsWith(SENSOR);
    }

    public boolean isStopCommand(String inputLine) {
        return inputLine.startsWith(STOP);
    }

    public boolean isRestartCommand(String inputLine) {
        return inputLine.startsWith(RESTART);
    }

    public boolean isCloseCommand(String inputLine) {
        return inputLine.startsWith(CLOSE);
    }

    public void close() throws IOException {
        inStream.close();
        outStream.close();
    }

}


class TTWPAction implements TankAction {
    float move = 0.0f;
    float turn = 0.0f;
    float aim = 0.0f;
    boolean fire = false;

    public void move(float f) {
        move = f;
    }

    public void turn(float f) {
        turn = f;
    }

    public void aim(float f) {
        aim = f;
    }

    public void fire() {
        fire = true;
    }

}

class TTWPException extends Exception {}

