/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * AI Tank
 *
 * Physical tank, sensory and movement systems.
 * Intelligence is held in a brain object.
 *
 * Author  Matthew Caryl
 * Created 30.10.96
 *
 * Although under copywrite to the author (Matthew Caryl)
 * this code can be copied and modified for non-commercial
 * purposes as long as any derivatives contain this condition.
 */

package ToyTank;

import java.awt.Graphics;
import java.awt.Color;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.lang.Math;
import ADT.Physical;
import ADT.Random;

public final class Tank implements TankAction, TankDescription, Physical {

    // Global tank data
    static int SHELLSPEED = 16;
    static int STARTTIME = 800;
    static float TANKSPEED = 2f;
    static float TURRETSPEED = 4f;

    // time stolen from other tank if shell hits
    static int TAKETIME = 8;
    static int GAINTIME = 8;//
    static int TICKTIME = 1;

    // cost of firing a shell
    static int FIRETIME = 1;

    // Maths routines
    static final float PI = (float) (Math.PI);
    static final float TWOPI = (float) (2.0 * Math.PI);

    // 2002-05-24
    static final float OCLOCK = (float) (Math.PI * 2f / 12f);
    static final float ZERO = OCLOCK * 0f;
    static final float ONE = OCLOCK * 1f;
    static final float THREE = OCLOCK * 3f;
    static final float SIX = OCLOCK * 6f;
    static final float NINE = OCLOCK * 9f;


    // 2000-06-19
    static final float ONE_DEGREE = 0.017453292f;

    static final float sin(float r) {
        return (float) Math.sin(r);
    }

    static final float cos(float r) {
        return (float) Math.cos(r);
    }

    static final float atan(float r) {
        return (float) Math.atan(r);
    }

    static final float sqr(float n) {
        return n * n;
    }

    static final float sqrt(float n) {
        return (float) Math.sqrt(n);
    }

    //
    // Private interface of tank
    //

    // polygon outlines converted to polar coordinates for easy manipulation
    private static final int[] xBodyOutline = {
        9, 11, -11, -9
    };
    private static final int[] yBodyOutline = {
        20, -20, -20, 20
    };

    // polar coordinates
    private static float[] aBodyOutline;
    private static float[] rBodyOutline;
    private static final int[] xTurretOutline = {
        2, 2, -2, -2
    };
    private static final int[] yTurretOutline = {
        18, -3, -3, 18
    };

    // polar coordinates
    private static float[] aTurretOutline;
    private static float[] rTurretOutline;
    private static int radius = 0;

    // Calculate polar outline co-ordinates
    static {
        aBodyOutline = new float[4];
        rBodyOutline = new float[4];

        for (int i = 0; i < 4; i++) {
            aBodyOutline[i] = (float) atan((float) xBodyOutline[i]
                                           / (float) yBodyOutline[i]);

            if (yBodyOutline[i] < 0)
                aBodyOutline[i] += PI;

            rBodyOutline[i] = sqrt(sqr(xBodyOutline[i])
                                   + sqr(yBodyOutline[i]));
        }

        aTurretOutline = new float[4];
        rTurretOutline = new float[4];

        for (int i = 0; i < 4; i++) {
            aTurretOutline[i] = (float) atan((float) xTurretOutline[i]
                                             / (float) yTurretOutline[i]);

            if (yTurretOutline[i] < 0)
                aTurretOutline[i] += PI;

            rTurretOutline[i] = sqrt(sqr(xTurretOutline[i])
                                     + sqr(yTurretOutline[i]));
        }

        float r = 0;

        for (int i = 0; i < 4; i++) {
            if (rBodyOutline[i] > r)
                r = rBodyOutline[i];

            if (rTurretOutline[i] > r)
                r = rTurretOutline[i];
        }

        radius = (int) r + 2;
    }

    // individual tank data
    private World world;
    private float x, y;
    private float heading;
    private float turret = 0;
    private Brain brain;
    private int time = STARTTIME;
    private float physicalCollisionAngle;
    private boolean physicalCollision = false;
    private float oldX, oldY;
    private float oldHeading;
    private float oldTurret;
    private int survival = 0;
    private int gap = 0;

    // 2001-05-25 use wide sensor for uniformity
    private TankWideSensor tankWideSensor;

    // individual tank display data
    private int[][][] outline = new int[2][2][5];
    private boolean dirtyOutline = true;
    private Rectangle bounds = new Rectangle();
    private boolean dirtyBound = true;
    private Color color;
    private String name = "<Unknown>";

    // generate attractive random colour
    private Color randomColor() {
        Random random = world.random();

        int[] colorlist = {
            random.range(64) + 192, random.range(64) + 128,
            random.range(64) + 64
        };

        for (int i = 3; i > 0; i--) {
            int p = random.range(3);
            int q = random.range(3);
            int temporary;

            temporary = colorlist[p];
            colorlist[p] = colorlist[q];
            colorlist[q] = temporary;
        }

        return new Color(colorlist[0], colorlist[1], colorlist[2]);
    }

    // remember previous position to return to after a collision
    private void remember() {
        oldX = x;
        oldY = y;
        oldHeading = heading;
        oldTurret = turret;
    }

    // is first angle less than second
    private boolean less(float first, float second) {
        if (first < second)
            return first + PI > second;
        else
            return first > second + PI;
    }

    // ensure angle is between [0..TWOPI]
    private float normalise(float angle) {
        if (angle < 0)
            return angle + TWOPI;
        else if (angle > TWOPI)
            return angle - TWOPI;
        else
            return angle;
    }

    // 2001-05-24
    private String bytes2String(byte[] bytes) {

        StringBuffer strb = new StringBuffer();

        for (int i = 0; i < bytes.length; i++)
            strb.append(bytes[i]);

        return strb.toString();
    }


    //
    // Public interface of tank
    //

    // initialise tank, with required brain
    public Tank(World w) {
        world = w;

        if (w.BRAIN_CLASS == Classifier.class) {
            brain = new Classifier(w.random());
        }
        else {
            try {
                brain = (Brain) w.BRAIN_CLASS.newInstance();
            }
            catch (Exception e) {
                throw new RuntimeException("Problem in brain creation: " + e);
            }
        }

        color = randomColor();

        teleport();
    }

    // initialise tank
    public Tank(World w, Brain b) {
        world = w;
        brain = b;
        color = randomColor();

        teleport();
    }

    public void setColor(Color c) {
        color = c;
    }

    public Color getColor() {
        return color;
    }

    public void setName(String n) {
        name = n;
    }

    public String getName() {
        return name;
    }

    public void setSurvival(int ticks) {
        survival = ticks;
    }

    public int getSurvival() {
        return survival;
    }

    public int updateGap(int best) {
        gap += best - survival;

        return gap;
    }

    public int getGap() {
        return gap;
    }

    // do NOT mutate tank
    public Tank breed() {
        turret = 0;
        time = STARTTIME;
        survival = 0;

        brain.restart();

        return this;
    }

    // new crossbreed tank
    // -- not implemented --
    public Tank breed(Tank mate) {
        throw new RuntimeException("Not implemented.");
    }

    // random position and angle
    public void teleport() {
        set(world.random().range(world.width()),
            world.random().range(world.height()));

        heading = (float) world.random().range(360) * PI / 180f;
    }

    // report of shell hitting wall
    public void hit(Shell s) {
        world.subPhysical((Physical) s);
    }

    // report of shell hitting physical object
    public void hit(Shell s, Physical p) {
        world.subPhysical((Physical) s);

        if (p instanceof Tank) {
            ((Tank) p).time -= TAKETIME;
            time += GAINTIME;
        }
        else
            ;   // do nothing for the moment
    }

    public int getResidualTime() {
        return time;
    }

    public void setResidualTime(int t) {
        time = t;
    }

    public World getWorld() {
        return world;
    }

    //
    // Public interface of physical
    //

    // retrieve outlines of tank components
    public int[][][] outline() {
        if (dirtyOutline) {
            for (int i = 0; i < 4; i++) {
                outline[0][0][i] = (int) (x
                                          + rBodyOutline[i]
                                            * sin(heading + aBodyOutline[i]));
                outline[0][1][i] = (int) (y
                                          - rBodyOutline[i]
                                            * cos(heading + aBodyOutline[i]));
            }

            outline[0][0][4] = outline[0][0][0];
            outline[0][1][4] = outline[0][1][0];

            for (int i = 0; i < 4; i++) {
                outline[1][0][i] = (int) (x
                                          + rTurretOutline[i]
                                            * sin(heading + turret
                                                  + aTurretOutline[i]));
                outline[1][1][i] = (int) (y
                                          - rTurretOutline[i]
                                            * cos(heading + turret
                                                  + aTurretOutline[i]));
            }

            outline[1][0][4] = outline[1][0][0];
            outline[1][1][4] = outline[1][1][0];
            dirtyOutline = false;
        }

        return outline;
    }

    // retrieve bounding rectangle of tank components
    public Rectangle bounds() {
        if (dirtyBound) {
            int minx, maxx, miny, maxy;
            int tx, ty;

            minx = maxx = (int) (x
                                 + rBodyOutline[0]
                                   * sin(heading + aBodyOutline[0]));
            miny = maxy = (int) (y
                                 - rBodyOutline[0]
                                   * cos(heading + aBodyOutline[0]));

            for (int i = 1; i < 4; i++) {
                tx = (int) (x
                            + rBodyOutline[i]
                              * sin(heading + aBodyOutline[i]));
                ty = (int) (y
                            - rBodyOutline[i]
                              * cos(heading + aBodyOutline[i]));

                if (tx < minx)
                    tx = minx;
                else if (tx > maxx)
                    tx = maxx;

                if (ty < miny)
                    ty = miny;
                else if (ty < maxy)
                    ty = maxy;
            }

            bounds.setBounds(minx, miny, maxx - minx, maxy - miny);

            dirtyBound = false;
        }

        return bounds;
    }

    // retrieve the maximum radius of the tank
    public int radius() {
        return radius;
    }

    public int x() {
        return (int) x;
    }

    public int y() {
        return (int) y;
    }

    // set coordinates, remember previous
    public void set(int X, int Y) {
        remember();

        x = X;
        y = Y;
        dirtyOutline = true;
        dirtyBound = true;
    }

    // has tank moved from remembered location
    public boolean moved() {
        return x != oldX || y != oldY || heading != oldHeading
               || turret != oldTurret;
    }

    // gather sensory information
    // 2001-05-25 use wide sensor for uniformity
    public void sensor() {

        if (tankWideSensor == null) {
            tankWideSensor = new TankWideSensor(this);
        }

        tankWideSensor.updateViews();

        TTWPSensor s =
            new TTWPSensor(String.valueOf(tankWideSensor.getEnergy()),
                           String.valueOf(tankWideSensor.getTurretAim()),
                           bytes2String(tankWideSensor.getCollisionView()),
                           bytes2String(tankWideSensor.getShortRangeView()),
                           bytes2String(tankWideSensor.getLongRangeView()));

        try {
            brain.sensor(s);
        }
        catch (Exception e) {

            // e.printStackTrace();
            // Out of game
            time = -1;
        }

    }

    // take action
    public void action() {
        remember();

        try {
            brain.action(this);
        }
        catch (Exception e) {

            // e.printStackTrace();
            // Out of game
            time = -1;
        }

        time -= TICKTIME;

        if (time < 0)
            world.subPhysical(this);

            // Reset

        physicalCollision = false;
    }

    // sensory and action
    public void tick() {
        sensor();
        action();
    }

    // revert tank back to its old position, wall collision
    public void collision(boolean first) {

        // 2002-05-24 pmarco
        // There is a subtle bug in Matthew's code that has been impossible
        // to find (at leaset in a reasonable time).
        // Symptom: sometimes false collision are reported.
        // It turned that the only way out is to verify
        if (!verifyCollision())
            return;

            // work out collision angle

        {
            int X = (int) oldX;
            int Y = (int) oldY;

            if (X != x)
                physicalCollisionAngle = atan((float) (X - x)
                                              / (float) (y - Y));
            else
                physicalCollisionAngle = Y < y ? 0 : PI;

            if (y < Y)
                physicalCollisionAngle += PI;

            physicalCollisionAngle += PI;

            physicalCollisionAngle = normalise(physicalCollisionAngle);
            physicalCollision = true;
        }

        x = oldX;
        y = oldY;
        heading = oldHeading;
        turret = oldTurret;
        dirtyOutline = true;
        dirtyBound = true;
    }

    // Revert tank back to its old position, physical collision
    public void collision(Physical p, boolean first) {

        // 2000-06-09 pmarco
        // Reverting tank back to its old position after a shell hit
        // makes it helpless against a tight sequence of incoming shots
        // Furthemore, having each shell hit detected as a collision may
        // disturb tank guidance
        if (p instanceof Shell)
            return;

        x = oldX;
        y = oldY;
        heading = oldHeading;
        turret = oldTurret;
        dirtyOutline = true;
        dirtyBound = true;

        // work out collision angle
        {
            int X = p.x();
            int Y = p.y();
            int x = x();
            int y = y();

            if (X != x)
                physicalCollisionAngle = atan((float) (X - x)
                                              / (float) (y - Y));
            else
                physicalCollisionAngle = Y < y ? 0 : PI;

            if (y < Y)
                physicalCollisionAngle += PI;

            physicalCollisionAngle = normalise(physicalCollisionAngle);
            physicalCollision = true;

        }
    }

    // draw new outline in tank colour
    public void paint(Graphics g) {
        g.setColor(color);
        outline();

        for (int i = outline.length - 1; i >= 0; i--)
            g.drawPolygon(outline[i][0], outline[i][1], outline[i][0].length);
    }

    // draw current outline in current colour
    public void erase(Graphics g) {
        for (int i = outline.length - 1; i >= 0; i--)
            g.drawPolygon(outline[i][0], outline[i][1], outline[i][0].length);
    }

    public void stop() {
        brain.stop();
    }

    public void close() {
        brain.close(getGap());
    }

    //
    // Public interface of tank sensor
    //

    // tank heading
    public float heading() {
        return heading;
    }

    // turret offset
    public float turret() {
        return turret;
    }

    // turret offset
    public int energy() {
        return time;
    }

    // any tanks visible between start and finish angle relative
    // to the turret within distance
    public boolean turretRadar(float start, float finish, float range) {
        int distance = (int) (radius * range);

        return world.searchAngle(this, heading + turret + start,
                                 heading + turret + finish, distance);
    }

    // any tanks visible between start and finish angle relative
    // to the tank within distance
    public boolean tankRadar(float start, float finish, float range) {
        int distance = (int) (radius * range);

        return world.searchAngle(this, heading + start, heading + finish,
                                 distance);
    }

    // did physical collision occur in this view field
    public boolean collisionSensor(float start, float finish) {
        if (!physicalCollision)
            return false;

        float from = normalise(start + heading);
        float to = normalise(finish + heading);

        return (less(from, physicalCollisionAngle)
                && less(physicalCollisionAngle, to));
    }

    // did wall collision occur in this view field
    public boolean obstacleSensor(float start, float finish, float range) {
        float distance = range;

        int x1 = (int) (x + sin(heading + start) * distance);
        int y1 = (int) (y - cos(heading + start) * distance);
        int x2 = (int) (x + sin(heading + finish) * distance);
        int y2 = (int) (y - cos(heading + finish) * distance);

        return world.checkLine((int) x, (int) y, x2, y2);
    }

    //
    // Public interface of tank action
    //

    // turn tank by angle
    public void turn(float angle) {

        // 2000-06-19 enforce limits
        angle = Math.max(-Tank.TANKSPEED * ONE_DEGREE,
                         Math.min(angle, Tank.TANKSPEED * ONE_DEGREE));

        heading += angle;
        heading = normalise(heading);

        dirtyOutline = true;
        dirtyBound = true;
    }

    // move distance forward
    public void move(float distance) {

        // 2000-06-19 enforce limits
        distance = Math.max(-Tank.TANKSPEED,
                            Math.min(distance, Tank.TANKSPEED));

        x += sin(heading) * distance;
        y -= cos(heading) * distance;
        dirtyOutline = true;
        dirtyBound = true;
    }

    // turn turret by angle
    public void aim(float angle) {

        // 2000-06-19 enforce limits
        angle = Math.max(-Tank.TURRETSPEED * ONE_DEGREE,
                         Math.min(angle, Tank.TURRETSPEED * ONE_DEGREE));

        turret += angle;
        turret = normalise(turret);

        dirtyOutline = true;
        dirtyBound = true;
    }

    // fire gun
    public void fire() {
        float direction = heading + turret;
        float distance = radius + Shell.radius;
        Shell s;

        s = new Shell(this, (int) (x + sin(direction) * distance),
                      (int) (y - cos(direction) * distance), direction,
                      SHELLSPEED, color);

        world.addPhysical((Physical) s);

        time -= FIRETIME;
    }

    // Bug fixing
    private boolean verifyCollision() {

        int range = TankWideSensor.SHORT_RANGE_DISTANCE + radius - 3;

        // Make sure that there is actually something around
        return (obstacleSensor(ZERO, THREE, range)
                || obstacleSensor(THREE, SIX, range)
                || obstacleSensor(SIX, NINE, range)
                || obstacleSensor(-THREE, ZERO, range));
    }

}

