/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * AI Tank
 *
 * Augmented sensory facility. Intended for compliancy with TTWP
 *
 * pmarco@ipvvis.unipv.it
 */

package ToyTank;

import java.lang.Math;
import ADT.Physical;

public class TankWideSensor {

    static final float PI = (float) (Math.PI);
    static final float TWOPI = (float) (2.0 * Math.PI);

    static final int COLLISION_SECTORS = 6;

    static final int SHORT_RANGE_DISTANCE = 8;
    static final int SHORT_RANGE_SECTORS = 12;

    static final int LONG_RANGE_DISTANCE = 512;
    static final int LONG_RANGE_START = 16;
    static final int LONG_RANGE_SEGMENTS = 6;
    static final int LONG_RANGE_SECTORS = 6;

    static final int OBSTACLE_COLOUR = 1;
    static final int PHYSICAL_COLOUR = 2;

    private static final float[] longRangeSectors = {
        -PI / 3, -PI / 6, -PI / 18, 0, PI / 18, PI / 6, PI / 3
    };

    private static final float[] longRangeSegments = {
        16, 32, 64, 128, 256, 512
    };

    private Tank tank;
    private World world;
    private float[] absRangeSectors;
    private byte[] collisionView;
    private byte[] shortRangeView;
    private byte[] longRangeView;

    static {

        // Use normalised angles everywhere
        for (int i = 0; i < longRangeSectors.length; i++)
            longRangeSectors[i] = normalise(longRangeSectors[i]);
    }

    public TankWideSensor(Tank t) {
        tank = t;
        world = tank.getWorld();
        absRangeSectors = new float[longRangeSectors.length];
        collisionView = new byte[COLLISION_SECTORS];
        shortRangeView = new byte[SHORT_RANGE_SECTORS];
        longRangeView = new byte[LONG_RANGE_SEGMENTS * LONG_RANGE_SECTORS];
    }

    // TTWP Sensory data
    public int getEnergy() {
        return tank.getResidualTime();
    }

    public float getTurretAim() {
        return tank.turret();
    }

    // Do this before each reading
    public void updateViews() {
        privateUpdateViews();
    }

    public byte[] getCollisionView() {
        return collisionView;
    }

    public byte[] getShortRangeView() {
        return shortRangeView;
    }

    public byte[] getLongRangeView() {
        return longRangeView;
    }

    // Maths routines
    static final float sin(float r) {
        return (float) Math.sin(r);
    }

    static final float cos(float r) {
        return (float) Math.cos(r);
    }

    static final float atan(float r) {
        return (float) Math.atan(r);
    }

    static final float sqr(float n) {
        return n * n;
    }

    static final float sqrt(float n) {
        return (float) Math.sqrt(n);
    }


    private void privateUpdateViews() {
        int x = tank.x();
        int y = tank.y();
        int radius = tank.radius();
        float heading = tank.heading();
        float viewing = normalise(heading + tank.turret());
        float shortRangeSector = TWOPI / SHORT_RANGE_SECTORS;
        float collisionSector = TWOPI / COLLISION_SECTORS;
        float angle;
        float distance;
        int X, X1;
        int Y, X2;

        // Reset views
        resetViews();

        // Update absolute range sectors
        for (int i = 0; i < longRangeSectors.length; i++)
            absRangeSectors[i] = normalise(longRangeSectors[i] + viewing);

        // Go through physicals
        Physical[] physicals = world.getCurrentPhysicals();

        for (int p = physicals.length - 1; p >= 0; p--) {
            if (physicals[p] == tank)

                // Skip: it's me
                continue;

            if (physicals[p] instanceof Shell)

                // A shell is too small and fast to be visible
                continue;

                // Compute distance & (absolute) angle

            X = physicals[p].x();
            Y = physicals[p].y();
            distance = sqrt(sqr(X - x) + sqr(Y - y)) - 2 * radius;

            if (X != x)
                angle = atan((float) (X - x) / (float) (y - Y));
            else
                angle = Y < y ? 0 : PI;

            if (y < Y)
                angle += PI;

            angle = normalise(angle);

            if (distance <= LONG_RANGE_DISTANCE
                    && less(absRangeSectors[0], angle)
                    && less(angle,
                            absRangeSectors[absRangeSectors.length - 1])) {

                // Check line of sight
                X = (int) (x + sin(angle) * distance);
                Y = (int) (y - cos(angle) * distance);

                if (world.checkLine((int) x, (int) y, X, Y))
                    continue;

                    // Find range

                for (int i = 0; i < longRangeSegments.length; i++) {
                    if (distance <= longRangeSegments[i]) {

                        // Find sector
                        for (int j = 0; j < LONG_RANGE_SECTORS; j++) {
                            if (less(absRangeSectors[j], angle)
                                    && less(angle, absRangeSectors[j + 1])) {

                                // That is, update view
                                longRangeView[i * LONG_RANGE_SECTORS + j] |=
                                    PHYSICAL_COLOUR;

                                break;
                            }
                        }

                        break;
                    }
                }
            }

            if (distance <= SHORT_RANGE_DISTANCE) {

                // Ok, insert in short range view
                float shortSight = heading;

                for (int j = 0; j < SHORT_RANGE_SECTORS; j++) {
                    if (less(shortSight, angle)
                            && less(angle, shortSight + shortRangeSector)) {

                        // That is, update view
                        shortRangeView[j] |= PHYSICAL_COLOUR;

                        break;
                    }

                    shortSight = normalise(shortSight + shortRangeSector);
                }
            }

        }   // Done with physicals

        // Go through obstacles

        // Update long range view
        for (int i = 0; i < LONG_RANGE_SECTORS; i++) {

            // Awful, but necessary
            angle =
                normalise(absRangeSectors[i]
                          + normalise((absRangeSectors[i + 1] - absRangeSectors[i])
                                      / 2));

            for (int j = 0; j < LONG_RANGE_SEGMENTS; j++) {
                X = (int) (x + sin(angle) * (longRangeSegments[j] + radius));
                Y = (int) (y - cos(angle) * (longRangeSegments[j] + radius));

                if (world.checkLine((int) x, (int) y, X, Y)) {
                    longRangeView[j * LONG_RANGE_SECTORS + i] |=
                        OBSTACLE_COLOUR;

                    // Anything beyond that is assumed invisible
                    break;
                }
            }
        }

        // Update short range view
        angle = normalise(heading + (shortRangeSector / 2));

        for (int i = 0; i < SHORT_RANGE_SECTORS; i++) {
            X = (int) (x + sin(angle) * (SHORT_RANGE_DISTANCE + radius));
            Y = (int) (y - cos(angle) * (SHORT_RANGE_DISTANCE + radius));

            if (world.checkLine((int) x, (int) y, X, Y))
                shortRangeView[i] |= OBSTACLE_COLOUR;

            angle = normalise(angle + shortRangeSector);
        }

        // Update collision view
        angle = 0;

        for (int i = 0; i < COLLISION_SECTORS; i++) {

            if (tank.collisionSensor(angle, angle + collisionSector))
                collisionView[i] = 1;

            angle = normalise(angle + collisionSector);
        }
    }

    // Reset views to null values
    private void resetViews() {
        for (int i = 0; i < longRangeView.length; i++)
            longRangeView[i] = 0;

        for (int i = 0; i < shortRangeView.length; i++)
            shortRangeView[i] = 0;

        for (int i = 0; i < collisionView.length; i++)
            collisionView[i] = 0;
    }


    // is first angle less than second
    private static boolean less(float first, float second) {
        if (first < second)
            return first + PI > second;
        else
            return first > second + PI;
    }

    // ensure angle is between [0..TWOPI]
    private static float normalise(float angle) {
        if (angle < 0)
            return angle + TWOPI;
        else if (angle > TWOPI)
            return angle - TWOPI;
        else
            return angle;
    }

}

