/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package evision.ec.np;

import java.util.*;

import ec.*;
import ec.util.*;


/**
 * Title:
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author
 * @version 1.0
 */

/**
 * An ADF is a NPNode which implements an "Automatically Defined Function",
 * similar to as described in Koza II.
 * *
 * <p>When an ADF is evaluated, it first evaluates all of its links
 * and stores away their results.  It then makeas a copy and evaluates its
 * associated graph. In the associated graph may exist one or more
 * <i>ADF Argument Terminals</i>, defined by the ADFArgument class.
 * These terminal nodes are associated with a single number which represents the
 * "argument" in the original ADF which evaluated their graph.
 * When an Argument Terminal is evaluated, it returns the stored result for that
 * link number in the parent ADF.
 * Ultimately, when the associated graph completes its evaluation, the ADF
 * returns that value.
 * *
 * <p>Obviously, if you have Argument Terminals in a graph, that graph must
 * be only callable by ADFs and ADMs, otherwise the Argument Terminals
 * won't have anything to return.  Furthermore, you must make sure that
 * you don't have an Argument Terminal in a graph whose number is higher
 * than the smallest arity (number of arguments) of a calling ADF or ADM.
 *
 *
 *
 * <p><b>Parameters</b><br>
 * <table>
 * <tr><td valign=top><i>base</i>.<tt>graph</tt><br>
 * <font size=-1>int &gt;= 0</font></td>
 * <td valign=top>(The "associated graph" of the ADF)</td></tr>
 * <tr><td valign=top><i>base</i>.<tt>name</tt><br>
 * <font size=-1>String, can be undefined</font></td>
 * <td valign=top>(A simple "name" of the ADF to distinguish it from other ADF
 * functions in your function set.  Use only letters, numbers, hyphens, and
 * underscores.  Lowercase is best.)</td></tr>
 * </table>
 * <p><b>Default Base</b><br>
 * gp.adf
 */

public class ADF extends NPNode {
    static final long serialVersionUID = -7125355148297077003L;

    public static final String P_ADF = "adf";
    public static final String P_CONTEXT = "context";
    public static final String P_ASSOCIATEDGRAPH = "graph";
    public static final String P_FUNCTIONNAME = "name";

    /**
     * The ADF's context prototype
     */
    public ADFContext context_proto;

    /**
     * The ADF's associated graph
     */
    public int associatedGraph;

    /**
     * The cloned&cached graph for execution
     */
    public transient NPGraph graph;

    /**
     * The context references for execution
     */
    public transient ADFContext context;

    /**
     * The "function name" of the ADF, to distinguish it from other ADF
     * functions you might provide.
     */
    public String functionName;

    public Parameter defaultBase() {
        return NPDefaults.base().push(P_ADF);
    }

    /**
     * Returns functionName.hashCode() + class.hashCode() + associatedGraph.  Hope
     * that's reasonably random.
     */

    public int nodeHashCode() {
        return (this.getClass().hashCode() + functionName.hashCode()
                + associatedGraph);
    }

    /**
     * Determines node equality by comparing the class, associated graph, and
     * function name of the nodes.
     */
    public boolean nodeEquals(final NPNode node) {
        if (!this.getClass().equals(node.getClass())
                || links.length != node.links.length)
            return false;

        ADF adf = (ADF) node;

        return (associatedGraph == adf.associatedGraph
                && functionName.equals(adf.functionName));
    }

    /**
     * Checks type-compatibility constraints between the ADF, its argument
     * terminals, and the graph type of its associated graph, and also checks
     * to make sure the graph exists, there aren't invalid argument terminals
     * in it, and there are sufficient argument terminals (a warning).  Whew!
     */
    public void checkConstraints(final EvolutionState state, final int graph,
                                 final NPIndividual typicalIndividual,
                                 final Parameter individualBase) {
        super.checkConstraints(state, graph, typicalIndividual,
                               individualBase);

        // does the associated graph exist?

        if (associatedGraph < 0
                || associatedGraph > typicalIndividual.graphs.length)
            state.output.error("The node " + toStringForError()
                               + " of individual " + individualBase
                               + " must have an associated graph that is >= 0 and < "
                               + typicalIndividual.graphs.length);
        else {

            // is the associated graph of the correct type?  Issue an error.

            if (!constraints().valuetype.compatibleWith(typicalIndividual.graphs[associatedGraph].constraints().graphtype))
                state.output.error("The return type of the node "
                                   + toStringForError() + " of individual "
                                   + individualBase
                                   + "is not type-compatible with the graph type of its associated graph.");

            NPFuncInfo[][] funcs =
    typicalIndividual.graphs[associatedGraph].constraints().functionset.nodes;

            boolean valid[] = new boolean[links.length];

            for (int w = 0; w < funcs.length; w++) {

                // does the graph's function set have argument terminals
                // that are beyond what I can provide?  (issue an error)

                NPFuncInfo[] gpfi = funcs[w];

                for (int x = 0; x < gpfi.length; x++)
                    if (gpfi[x].node instanceof ADFArgument) {
                        int arg = ((ADFArgument) (gpfi[x].node)).argument;

                        if (arg >= links.length)// uh oh
                            state.output.error("The node "
                                               + toStringForError()
                                               + " in individual "
                                               + individualBase
                                               + " would call its associated graph, "
                                               + "which has an argument terminal "
                                               + "with an argument number ("
                                               + arg
                                               + ") >= the ADF/ADM's arity ("
                                               + links.length
                                               + ").  The argument terminal in question is "
                                               + gpfi[x].node.toStringForError());
                        else {
                            if (valid[arg])     // got one already
                                state.output.warning("There exists more than one "
                                                     + "Argument terminal for argument #"
                                                     + arg + " for the node "
                                                     + toStringForError()
                                                     + " in individual "
                                                     + individualBase);

                            else
                                valid[arg] = true;

                            // is the argument terminal of the correct return type?  Issue an error.
                            if (!gpfi[x].node.constraints().valuetype.compatibleWith(constraints().linktypes[arg]))
                                state.output.error("The node "
                                                   + toStringForError()
                                                   + " in individual "
                                                   + individualBase
                                                   + " would call its associated graph, "
                                                   + "which has an argument terminal "
                                                   + "which is not type-compatible "
                                                   + "with the related argument position "
                                                   + "of the ADF/ADM.  The argument terminal in question is "
                                                   + gpfi[x].node.toStringForError());
                        }
                    }
            }

            // does the graph's function set have fewer argument terminals
            // than I can provide? (issue a warning)

            for (int x = 0; x < links.length; x++)
                if (!valid[x])
                    state.output.warning("There is no argument terminal for argument #"
                                         + x + " for the node "
                                         + toStringForError()
                                         + " in individual "
                                         + individualBase);

        }
    }

    public void setup(final EvolutionState state, final Parameter base) {

        // we don't know our name yet, (used in toStringForError(),
        // which is used in NPNode's setup(...) method),
        // so WE load parameters before our parent does.

        Parameter def = defaultBase();

        functionName = state.parameters.getString(base.push(P_FUNCTIONNAME),
                                                  def.push(P_FUNCTIONNAME));

        if (functionName == null) {
            state.output.warning("ADF node has no function name.  Using a blank name.",
                                 base.push(P_FUNCTIONNAME),
                                 def.push(P_FUNCTIONNAME));

            functionName = "";
        }

        associatedGraph =
            state.parameters.getInt(base.push(P_ASSOCIATEDGRAPH),
                                    def.push(P_FUNCTIONNAME), 0);

        if (associatedGraph < 0)
            state.output.fatal("ADF node must have a positive-numbered associated graph.",
                               base.push(P_ASSOCIATEDGRAPH),
                               def.push(P_FUNCTIONNAME));

            // now we let our parent set up.

        super.setup(state, base);

        // load our prototype

        Parameter p = base.push(P_CONTEXT);
        Parameter d = defaultBase().push(P_CONTEXT);

        context_proto =
            (ADFContext) (state.parameters.getInstanceForParameterEq(p, d,
                ADFContext.class));

        context_proto.setup(state, p);
    }

    public void resetMemory(final EvolutionState state, final int thread) {

        super.resetMemory(state, thread);

        // Propagate to graph
        if (graph != null) {
            graph.resetGraph(state, thread);
        }
    }

    public String toString() {
        return "ADF" + functionName + "[" + associatedGraph + "]";
    }

    public void eval(final EvolutionState state, final int thread,
                     final NPData input, final ADFContext context,
                     final NPMemory newMemory, final NPIndividual individual,
                     final Problem problem) {

        // Make sure that graph is set
        provideAssociatedGraph(state, thread, individual);

        // Make sure that context is set
        provideContext();

        // evaluate the top of the associatedGraph
        individual.graphs[associatedGraph].eval(state, thread, input, context,
                                                individual, problem);

    }

    public void provideContext() {
        if (context == null) {
            context = (ADFContext) (context_proto.protoCloneSimple());
            context.prepare(this);
        }
    }

    public void provideAssociatedGraph(final EvolutionState state,
                                       final int thread,
                                       final NPIndividual individual) {
        if (graph == null) {
            graph =
                (NPGraph) individual.graphs[associatedGraph].protoCloneSimple();

            graph.resetGraph(state, thread);
        }
    }

    public Object protoClone() throws CloneNotSupportedException {
        ADF myobj = (ADF) (super.clone());

        // deep-cloned stuff
        myobj.context_proto = (ADFContext) (context_proto.protoClone());

        // Do not clone graph & context
        return myobj;
    }

}

