/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package evision.ec.np;

/**
 * Title:
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author
 * @version 1.0
 */

import java.io.*;
import java.util.*;

import ec.*;
import ec.util.*;

/**
 * NPFunctionSet is a Clique which represents a set of NPNode prototypes
 * forming a standard function set for forming certain trees in individuals.
 * NPFunctionSets instances have unique names with which they're referenced by
 * NPGraphConstraints objects indicating that they're used for certain trees.
 * NPFunctionSets store their NPNode Prototypes in three hashtables,
 * one for all nodes, one for nonterminals, and one for terminals.  Each
 * hashed item is an array of NPFuncInfo objects containing the NPNodes proper,
 * hashed by the return type of the NPNodes in the array.
 *
 * NPFunctionSets also contain prototypical NPFuncInfo nodes which they
 * clone to form their arrays.
 * <p><b>Parameters</b><br>
 * <table>
 * <tr><td valign=top><i>base</i>.<tt>name</tt><br>
 * <font size=-1>String</font></td>
 * <td valign=top>(name of function set.  Must be different from other function set instances)</td></tr>
 * <tr><td valign=top><i>base</i>.<tt>size</tt><br>
 * <font size=-1>int &gt;= 1</font></td>
 * <td valign=top>(number of functions in the function set)</td></tr>
 * <tr><td valign=top><i>base</i>.<tt>func.</tt><i>n</i><br>
 * <font size=-1>classname, inherits and != ec.gp.NPNode</font></td>
 * <td valign=top>(class of function node <i>n</i> in the set)</td></tr>
 * <tr><td valign=top><i>base</i>.<tt>info</tt><br>
 * <font size=-1>classname, inherits or == ec.gp.NPFuncInfo,<br>or nothing</font></td>
 * <td valign=top>(class of the prototypical NPFuncInfo object.  If nothing, this defaults to ec.gp.NPFuncInfo)</td></tr>
 * </table>
 * <p><b>Parameter bases</b><br>
 * <table>
 * <tr><td valign=top><i>base</i>.<tt>info</tt></td>
 * <td>p_funcinfo</td></tr>
 * <tr><td valign=top><i>base</i>.<tt>func.</tt><i>n</i></td>
 * <td>function node <i>n</i></td></tr>
 * </table>
 *
 */

public class NPFunctionSet implements Clique {
    static final long serialVersionUID = -8161928274224596786L;

    public final static String P_NAME = "name";
    public final static String P_SIZE = "size";
    public final static String P_FUNCINFO = "info";
    public final static String P_FUNC = "func";

    /**
     * Name of the NPFunctionSet
     */
    public String name;

    /**
     * The nodes that our NPGraph can use: arrays of nodes hashed by type.
     */
    public Hashtable nodes_h;

    /**
     * The nodes that our NPGraph can use: nodes[type][thenodes].
     */
    public NPFuncInfo[][] nodes;

    /**
     * The nonterminals our NPGraph can use: arrays of nonterminals hashed by type.
     */
    public Hashtable nonterminals_h;

    /**
     * The nonterminals our NPGraph can use: nonterminals[type][thenodes].
     */
    public NPFuncInfo[][] nonterminals;

    /**
     * The terminals our NPGraph can use: arrays of terminals hashed by type.
     */
    public Hashtable terminals_h;

    /**
     * The terminals our NPGraph can use: terminals[type][thenodes].
     */
    public NPFuncInfo[][] terminals;


    // some convenience methods which speed up various kinds
    // of mutation operators


    /**
     * Nodes == a given arity, that is: nodesByArity[type][arity][thenodes]
     */
    public NPFuncInfo[][][] nodesByArity;

    /**
     * Nonterminals <= a given arity, that is: nonterminalsUnderArity[type][arity][thenodes] --
     * this will be O(n^2).  Obviously, the number of nonterminals at arity slot 0 is 0.
     */
    public NPFuncInfo[][][] nonterminalsUnderArity;

    /**
     * Nonterminals >= a given arity, that is: nonterminalsOverArity[type][arity][thenodes] --
     * this will be O(n^2).  Obviously, the number of nonterminals at arity slot 0 is all the
     * nonterminals of that type.
     */
    public NPFuncInfo[][][] nonterminalsOverArity;


    /**
     * The prototypical NPFuncInfo node.
     */
    public NPFuncInfo p_funcinfo;

    /**
     * A global storage facility for all known NPFunctionSet objects.
     */
    public static Hashtable all;

    static {
        all = new Hashtable();
    }

    /**
     * Returns the name.
     */
    public String toString() {
        return name;
    }


    /**
     * Sets up all the NPFunctionSet, loading them from the parameter
     * file.  This must be called before anything is called which refers
     * to a type by name.
     */

    public static void setupFunctionSets(final EvolutionState state,
                                         final Parameter base) {
        state.output.message("Processing NP Function Sets");

        // How many NPFunctionSets do we have?
        int x = state.parameters.getInt(base.push(P_SIZE), null, 1);

        if (x <= 0)
            state.output.fatal("The number of NPFunctionSets must be at least 1.",
                               base.push(P_SIZE));

            // Load our FunctionSet

        for (int y = 0; y < x; y++) {
            NPFunctionSet c;

            // Figure the NPFunctionSet class
            if (state.parameters.exists(base.push("" + y)))
                c = (NPFunctionSet) (state.parameters.getInstanceForParameterEq(base.push(""
                        + y), null, NPFunctionSet.class));
            else {
                state.output.message("No NPFunctionSet specified, assuming the default class: ec.gp.NPFunctionSet for "
                                     + base.push("" + y));

                c = new NPFunctionSet();
            }

            c.setup(state, base.push("" + y));
        }
    }


    /**
     * Sets up the arrays based on the hashtables
     */

    public void postProcessFunctionSet() {
        nodes = new NPFuncInfo[nodes_h.size()][];
        terminals = new NPFuncInfo[terminals_h.size()][];
        nonterminals = new NPFuncInfo[nonterminals_h.size()][];

        Enumeration e = nodes_h.keys();

        while (e.hasMoreElements()) {
            NPType gpt = (NPType) (e.nextElement());
            NPFuncInfo[] gpfi = (NPFuncInfo[]) (nodes_h.get(gpt));

            nodes[gpt.type] = gpfi;
        }

        e = nonterminals_h.keys();

        while (e.hasMoreElements()) {
            NPType gpt = (NPType) (e.nextElement());
            NPFuncInfo[] gpfi = (NPFuncInfo[]) (nonterminals_h.get(gpt));

            nonterminals[gpt.type] = gpfi;
        }

        e = terminals_h.keys();

        while (e.hasMoreElements()) {
            NPType gpt = (NPType) (e.nextElement());
            NPFuncInfo[] gpfi = (NPFuncInfo[]) (terminals_h.get(gpt));

            terminals[gpt.type] = gpfi;
        }

        // set up arity-based arrays
        // first, determine the maximum arity
        int max_arity = 0;

        for (int x = 0; x < nodes.length; x++)
            for (int y = 0; y < nodes[x].length; y++)
                if (max_arity < nodes[x][y].node.links.length)
                    max_arity = nodes[x][y].node.links.length;

                    // next set up the == array

        nodesByArity = new NPFuncInfo[nodes.length][max_arity + 1][];

        for (int x = 0; x < nodes.length; x++)
            for (int a = 0; a <= max_arity; a++) {

                // how many nodes do we have?
                int num_of_a = 0;

                for (int y = 0; y < nodes[x].length; y++)
                    if (nodes[x][y].node.links.length == a)
                        num_of_a++;

                        // allocate and fill

                nodesByArity[x][a] = new NPFuncInfo[num_of_a];

                int cur_a = 0;

                for (int y = 0; y < nodes[x].length; y++)
                    if (nodes[x][y].node.links.length == a)
                        nodesByArity[x][a][cur_a++] = nodes[x][y];
            }

        // now set up the <= nonterminals array
        nonterminalsUnderArity =
            new NPFuncInfo[nonterminals.length][max_arity + 1][];

        for (int x = 0; x < nonterminals.length; x++)
            for (int a = 0; a <= max_arity; a++) {

                // how many nonterminals do we have?
                int num_of_a = 0;

                for (int y = 0; y < nonterminals[x].length; y++)
                    if (nonterminals[x][y].node.links.length <= a)
                        num_of_a++;

                        // allocate and fill

                nonterminalsUnderArity[x][a] = new NPFuncInfo[num_of_a];

                int cur_a = 0;

                for (int y = 0; y < nonterminals[x].length; y++)
                    if (nonterminals[x][y].node.links.length <= a)
                        nonterminalsUnderArity[x][a][cur_a++] =
                            nonterminals[x][y];
            }



        // now set up the >= nonterminals array
        nonterminalsOverArity =
            new NPFuncInfo[nonterminals.length][max_arity + 1][];

        for (int x = 0; x < nonterminals.length; x++)
            for (int a = 0; a <= max_arity; a++) {

                // how many nonterminals do we have?
                int num_of_a = 0;

                for (int y = 0; y < nonterminals[x].length; y++)
                    if (nonterminals[x][y].node.links.length >= a)
                        num_of_a++;

                        // allocate and fill

                nonterminalsOverArity[x][a] = new NPFuncInfo[num_of_a];

                int cur_a = 0;

                for (int y = 0; y < nonterminals[x].length; y++)
                    if (nonterminals[x][y].node.links.length >= a)
                        nonterminalsOverArity[x][a][cur_a++] =
                            nonterminals[x][y];
            }

    }




    /**
     * Must be done <i>after</i> NPType and NPNodeConstraints have been set up
     */

    public void setup(final EvolutionState state, final Parameter base) {

        // What's my name?
        name = state.parameters.getString(base.push(P_NAME), null);

        if (name == null)
            state.output.fatal("No name was given for this function set.",
                               base.push(P_NAME));

            // Register me

        NPFunctionSet old_functionset = (NPFunctionSet) (all.put(name, this));

        if (old_functionset != null)
            state.output.fatal("The NPFunctionSet \"" + name
                               + "\" has been defined multiple times.", base.push(P_NAME));

            // What's my function info class?

        if (state.parameters.exists(base.push(P_FUNCINFO)))
            p_funcinfo =
                (NPFuncInfo) (state.parameters.getInstanceForParameterEq(base.push(P_FUNCINFO),
                    null, NPFuncInfo.class));
        else {
            state.output.message("Using default NPFuncInfo class for NPFunctionSet "
                                 + name + ".");

            p_funcinfo = new NPFuncInfo();
        }

        p_funcinfo.setup(state, base.push(P_FUNCINFO));


        // How many functions do I have?
        int numFuncs = state.parameters.getInt(base.push(P_SIZE), null, 1);

        if (numFuncs < 1)
            state.output.error("The NPFunctionSet \"" + name
                               + "\" has no functions.", base.push(P_SIZE));

        Parameter p = base.push(P_FUNC);
        Vector tmp = new Vector();

        for (int x = 0; x < numFuncs; x++) {

            // load
            NPFuncInfo gpfi = (NPFuncInfo) (p_funcinfo.protoCloneSimple());
            Parameter pp = p.push("" + x);

            gpfi.node = (NPNode) (state.parameters.getInstanceForParameter(pp,
                    null, NPNode.class));

            gpfi.node.setup(state, pp);

            // NPNodes may well be memoryless
            if (gpfi.node.memory != null) {
                gpfi.node.memory.setup(state, pp);
            }

            // add to my collection
            tmp.addElement(gpfi);
        }

        // Make my hash tables
        nodes_h = new Hashtable();
        terminals_h = new Hashtable();
        nonterminals_h = new Hashtable();

        // Now set 'em up according to the types in NPType

        Enumeration e = NPType.all.elements();

        while (e.hasMoreElements()) {
            NPType typ = (NPType) (e.nextElement());

            // make vectors for the type.
            Vector nodes_v = new Vector();
            Vector terminals_v = new Vector();
            Vector nonterminals_v = new Vector();

            // add NPFuncInfos as appropriate to each vector
            Enumeration v = tmp.elements();

            while (v.hasMoreElements()) {
                NPFuncInfo i = (NPFuncInfo) (v.nextElement());

                if (typ.compatibleWith(i.node.constraints().valuetype)) {
                    nodes_v.addElement(i);

                    if (i.node.links.length == 0)
                        terminals_v.addElement(i);
                    else
                        nonterminals_v.addElement(i);
                }
            }

            // turn nodes_h' vectors into arrays
            NPFuncInfo[] ii = new NPFuncInfo[nodes_v.size()];

            nodes_v.copyInto(ii);
            nodes_h.put(typ, ii);

            // turn terminals_h' vectors into arrays
            ii = new NPFuncInfo[terminals_v.size()];

            terminals_v.copyInto(ii);
            terminals_h.put(typ, ii);

            // turn nonterminals_h' vectors into arrays
            ii = new NPFuncInfo[nonterminals_v.size()];

            nonterminals_v.copyInto(ii);
            nonterminals_h.put(typ, ii);

            // Post-arrange the items
            nodes_h = p_funcinfo.arrange(nodes_h);
            terminals_h = p_funcinfo.arrange(terminals_h);
            nonterminals_h = p_funcinfo.arrange(nonterminals_h);
        }

        // I don't check to see if the generation mechanism will be valid here
        // -- I check that in NPGraphConstraints, where I can do the weaker check
        // of going top-down through functions rather than making sure that every
        // single function has a compatible argument function (an unneccessary check)

        state.output.exitIfErrors();// because I promised when I called n.setup(...)

        // postprocess the function set
        postProcessFunctionSet();
    }


    /**
     * Returns the function set for a given name.
     * You must guarantee that after calling functionSetFor(...) one or
     * several times, you call state.output.exitIfErrors() once.
     */

    public static NPFunctionSet functionSetFor(final String name,
                                               final EvolutionState state) {
        NPFunctionSet set = (NPFunctionSet) (all.get(name));

        if (set == null)
            state.output.error("The NP function set \"" + name
                               + "\" could not be found.");

        return set;
    }


    public final void writeObject(ObjectOutputStream out) throws IOException {

        // this wastes an hashtable pointer, but what the heck.

        out.defaultWriteObject();
        out.writeObject(all);
    }

    public final void readObject(ObjectInputStream in)
            throws IOException, ClassNotFoundException {
        in.defaultReadObject();

        all = (Hashtable) (in.readObject());
    }

}

