/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package evision.ec.np;

import java.io.*;
import java.util.Vector;

import ec.*;
import ec.util.*;
import ec.EvolutionState;

/**
 * Title:
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author
 * @version 1.0
 */

/**
 * NPGraph holds a graph  of NPNodes.
 * NPGraph typically fill out an array held in a NPIndividual
 * (their "owner") and their roots are evaluated to evaluate a neural
 * programming graphs.
 *
 * NPGraph also have <i>constraints</i>, which are shared, and define items
 * shared among several NPGraph.
 *
 * <p><b>Parameters</b><br>
 * <table>
 * <tr><td valign=top><i>base</i>.<tt>tc</tt><br>
 * <font size=-1>String</font></td>
 * <td valign=top>(The graph's constraints)</td></tr>
 * </table>
 * <p><b>Default Base</b><br>
 * np.graph
 */

public class NPGraph implements NPNodeParent {
    static final long serialVersionUID = -5653777617927983249L;

    public static final String P_GRAPH = "graph";
    public static final String P_GRAPHCONSTRAINTS = "gc";
    public static final int NO_GRAPHNUM = -1;

    /**
     * the owner of the NPGraph
     */
    public NPIndividual owner;

    /**
     * the cache for new memory values
     */
    public NPNode[] nodes;

    /**
     * the cache for new memory values, do not serialize
     */
    public transient NPMemory[] cache;

    /**
     * constraints on the NPGraph  -- don't access the constraints through
     * this variable -- use the constraints() method instead, which will give
     * the actual constraints object.
     */
    public byte constraints;

    public final NPGraphConstraints constraints() {
        return NPGraphConstraints.constraints[constraints];
    }

    public Parameter defaultBase() {
        return NPDefaults.base().push(P_GRAPH);
    }

    /**
     * Returns true if I am "genetically" the same as graph.
     */
    public boolean graphEquals(NPGraph graph) {

        if (nodes.length != graph.nodes.length) {
            return false;
        }

        for (int i = 0; i < nodes.length; i++) {
            if (!nodes[i].nodeEquals(graph.nodes[i]))
                return false;
        }

        return true;
    }

    /**
     * Returns a hash code for comparing different NPGraphs.  In
     * general, two graphs which are graphEquals(...) should have the
     * same hash code.
     */
    public int graphHashCode() {
        if (nodes.length == 0) {
            return (this.getClass().hashCode());
        }

        int hash = nodes[0].nodeHashCode();

        for (int i = 1; i < nodes.length; i++) {

            // rotate hash and XOR
            hash = (hash << 1 | hash >>> 31) ^ nodes[1].nodeHashCode();
        }

        return hash;
    }

    // Mainly for debugging
    public void verifyGraph(boolean unconnectedNodes) {

        if (!unconnectedNodes && nodes.length > 1) {

            // Mark
            boolean[] marks = new boolean[nodes.length];

            for (int i = 0; i < marks.length; i++) {
                marks[i] = false;
            }

            for (int i = 0; i < nodes.length; i++) {

                for (int j = 0; j < nodes[i].links.length; j++) {
                    if (nodes[i].links[j] != nodes[i]) {
                        marks[nodes[i].links[j].position] = true;
                        marks[i] = true;
                    }
                }
            }

            // Verify
            for (int i = 0; i < nodes.length; i++) {
                if (marks[i] == false) {
                    throw new RuntimeException("Unconnected node");
                }
            }
        }


        for (int i = 0; i < nodes.length; i++) {

            if (nodes[i].owner != this) {
                throw new RuntimeException("Wrong node owner");
            }

            if (nodes[i].position != i) {
                throw new RuntimeException("Wrong node position");
            }

            for (int j = 0; j < nodes[i].links.length; j++) {
                if (nodes[i].links[j] != nodes[nodes[i].links[j].position]) {
                    throw new RuntimeException("Misplaced link");
                }
            }
        }
    }

    /**
     * Deep-clones the graph, i.e. graph and all nodes.
     */
    public Object protoClone() throws CloneNotSupportedException {
        NPGraph graph = (NPGraph) super.clone();

        // Cut cache, anyway
        graph.cache = null;

        // A prototype NPGraph may have no nodes
        if (nodes != null) {

            NPNode[] fromNodes = graph.nodes;

            graph.nodes = (NPNode[]) nodes.clone();

            // Clone nodes
            NPNode[] toNodes = graph.nodes;

            for (int i = 0; i < fromNodes.length; i++) {
                toNodes[i] = (NPNode) fromNodes[i].protoClone();
                toNodes[i].memory =
                    (NPMemory) fromNodes[i].memory.protoClone();

                fromNodes[i].memory.copyTo(toNodes[i].memory);

                toNodes[i].owner = graph;
            }

            // Fix links
            for (int i = 0; i < toNodes.length; i++) {
                for (int j = 0; j < toNodes[i].links.length; j++) {
                    toNodes[i].links[j] =
                        toNodes[fromNodes[i].links[j].position];
                }
            }
        }

        return graph;
    }

    public final Object protoCloneSimple() {
        try {
            return protoClone();
        }
        catch (CloneNotSupportedException e) {
            throw new InternalError();
        }   // never happens
     }

    /**
     * An expensive function which determines my graph number -- only
     * use for errors, etc. Returns ec.gp.NPGraph.NO_GRAPHNUM if the
     * graph number could not be
     * determined (might happen if it's not been assigned yet).
     */
    public int graphNumber() {
        if (owner == null)
            return NO_GRAPHNUM;

        if (owner.graphs == null)
            return NO_GRAPHNUM;

        for (int x = 0; x < owner.graphs.length; x++)
            if (owner.graphs[x] == this)
                return x;

        return NO_GRAPHNUM;
    }


    /**
     * Sets up a prototypical NPGraph with those features it shares with
     * other NPGraphs in its position in its NPIndividual, and nothhing more.
     * This must be called <i>after</i> the NPTypes and NPNodeConstraints
     * have been set up.  Presently they're set up in NPInitializer,
     * which gets called before this does, so we're safe.
     */
    public void setup(final EvolutionState state, final Parameter base) {
        Parameter def = defaultBase();

        // determine my constraints -- at this point, the constraints should have been loaded.
        String s = state.parameters.getString(base.push(P_GRAPHCONSTRAINTS),
                                              def.push(P_GRAPHCONSTRAINTS));

        if (s == null)
            state.output.fatal("No graph constraints are defined for the NPGraph "
                               + base + ".");
        else
            constraints = NPGraphConstraints.constraintsFor(s,
                    state).constraintNumber;

        state.output.exitIfErrors();// because I promised

        // we're not loading the nodes at this point
    }


    public void printGraphForHumans(final EvolutionState state,
                                    final int log, final int verbosity) {

        StringBuffer strb = new StringBuffer();

        for (int i = 0; i < nodes.length; i++) {
            strb.setLength(0);
            strb.append("[").append(i).append("] : ");
            strb.append(nodes[i].toStringForHumans()).append(" ");

            for (int j = 0; j < nodes[i].links.length; j++) {
                strb.append("[").append(nodes[i].links[j].position).append("]");
            }

            state.output.println(strb.toString(), verbosity, log);
        }
    }

    /**
     * Builds a new randomly-generated graph.
     */

    public void buildGraph(final EvolutionState state, final int thread)
            throws CloneNotSupportedException {

        nodes = constraints().init.newGraph(state, constraints().graphtype,
                                            thread, this,
                                            constraints().functionset, 0,
                                            NPGraphBuilder.NOSIZEGIVEN);
    }

    /**
     * Reset all node memories, e.g. after a previous execution
     */

    public void resetGraph(final EvolutionState state, final int thread) {

        for (int i = 0; i < nodes.length; i++) {
            nodes[i].resetMemory(state, thread);
        }
    }


    /**
     * Evaluation has to take place on the entire graph
     */

    public void eval(final EvolutionState state, final int thread,
                     final NPData input, final ADFContext context,
                     final NPIndividual individual, final Problem problem) {

        if (cache == null) {

            // Ok, setup cache
            try {
                cache = new NPMemory[nodes.length];

                for (int i = 0; i < nodes.length; i++) {

                    // NPNodes may well be stateless
                    if (nodes[i].memory != null) {
                        cache[i] = (NPMemory) nodes[i].memory.protoClone();
                    }
                }
            }
            catch (CloneNotSupportedException e) {
                throw new InternalError();
            }   // never happens
         }

        // Step 1: eval
        for (int i = 0; i < nodes.length; i++) {
            nodes[i].eval(state, thread, input, context, cache[i],
                          individual, problem);
        }

        // Step 2: update nodes
        for (int i = 0; i < nodes.length; i++) {
            cache[i].copyTo(nodes[i].memory);
        }

    }

    /**
     * Collecting all unconnected nodes from an NPNode array
     */

    public void collectUnconnectedNodes() {

        if (nodes.length == 1) {
            return;
        }

        // Mark
        boolean[] marks = new boolean[nodes.length];

        for (int i = 0; i < marks.length; i++) {
            marks[i] = false;
        }

        for (int i = 0; i < nodes.length; i++) {

            for (int j = 0; j < nodes[i].links.length; j++) {
                if (nodes[i].links[j] != nodes[i]) {
                    marks[nodes[i].links[j].position] = true;
                    marks[i] = true;
                }
            }
        }

        // Sweep
        Vector survivors = new Vector();

        for (int i = 0; i < nodes.length; i++) {
            if (marks[i] == true) {
                nodes[i].position = survivors.size();

                survivors.add(nodes[i]);
            }
        }

        NPNode[] newNodes = new NPNode[survivors.size()];

        survivors.copyInto(newNodes);

        nodes = newNodes;
    }

}

