/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package evision.ec.np;

import ec.*;
import ec.util.*;

import java.io.*;
import java.util.*;

/**
 * Title:
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author
 * @version 1.0
 */

/**
 * A NPGraphConstraints is a Clique which defines constraint information
 * common to many different NPGraph graphs, namely the graph type,
 * builder, and function set.  NPGraphConstraints have unique names
 * by which they are identified.
 *
 * <p>In adding new things to NPGraphConstraints, you should ask yourself
 * the following questions: first, is this something that takes up too
 * much memory to store in NPGraphs themseves?  second, is this something
 * that needs to be accessed very rapidly, so cannot be implemented as
 * a method call in a NPGraph?  third, can this be shared among different
 * NPGraphs?
 *
 * <p><b>Parameters</b><br>
 * <table>
 * <tr><td valign=top><i>base</i>.<tt>size</tt><br>
 * <font size=-1>int &gt;= 1</font></td>
 * <td valign=top>(number of graph constraints)</td></tr>
 * <tr><td valign=top><i>base.n</i>.<tt>name</tt><br>
 * <font size=-1>String</font></td>
 * <td valign=top>(name of graph constraint <i>n</i>)</td></tr>
 * <tr><td valign=top><i>base.n</i>.<tt>init</tt><br>
 * <font size=-1>classname, inherits and != ec.gp.NPNodeBuilder</font></td>
 * <td valign=top>(NP node builder for graph constraint <i>n</i>)</td></tr>
 * <tr><td valign=top><i>base.n</i>.<tt>returns</tt><br>
 * <font size=-1>String</font></td>
 * <td valign=top>(graph type for graph constraint <i>n</i>)</td></tr>
 * <tr><td valign=top><i>base.n</i>.<tt>fset</tt><br>
 * <font size=-1>String</font></td>
 * <td valign=top>(function set for graph constraint <i>n</i>)</td></tr>
 * </table>
 * @author Sean Luke
 * @version 1.0
 */

public class NPGraphConstraints implements Clique {
    static final long serialVersionUID = 3834944249381917346L;

    public static final int SIZE_OF_BYTE = 256;
    public final static String P_NAME = "name";
    public final static String P_SIZE = "size";
    public final static String P_INIT = "init";
    public static final String P_VALUE = "value";
    public static final String P_FUNCTIONSET = "fset";

    public String name;

    /**
     * The byte value of the constraints -- we can only have 256 of them
     */
    public byte constraintNumber;

    /**
     * The builder for the graph
     */
    public NPGraphBuilder init;

    /**
     * The type of of the graph
     */
    public NPType graphtype;

    /**
     * The function set for nodes in the graph
     */
    public NPFunctionSet functionset;

    /**
     * A reposiory of all the NPGraphConstraints in the system.
     */
    public static Hashtable all;
    public static NPGraphConstraints[] constraints;
    public static byte numConstraints;

    static {
        all = new Hashtable();
        constraints = new NPGraphConstraints[SIZE_OF_BYTE];
        numConstraints = 0;
    }

    public String toString() {
        return name;
    }



    /**
     * Sets up all the NPGraphConstraints, loading them from the parameter
     * file.  This must be called before anything is called which refers
     * to a type by name.
     */

    public static void setupConstraints(final EvolutionState state,
                                        final Parameter base) {
        state.output.message("Processing NP Graph Constraints");

        // How many NPGraphConstraints do we have?
        int x = state.parameters.getInt(base.push(P_SIZE), null, 1);

        if (x <= 0)
            state.output.fatal("The number of NP graph constraints must be at least 1.",
                               base.push(P_SIZE));

            // Load our constraints

        for (int y = 0; y < x; y++) {
            NPGraphConstraints c;

            // Figure the constraint class
            if (state.parameters.exists(base.push("" + y)))
                c = (NPGraphConstraints) (state.parameters.getInstanceForParameterEq(base.push(""
                        + y), null, NPGraphConstraints.class));
            else {
                state.output.message("No NP Graph Constraints specified, assuming the default class: ec.gp.NPGraphConstraints for "
                                     + base.push("" + y));

                c = new NPGraphConstraints();
            }

            c.setup(state, base.push("" + y));
        }

        // set our constraints array up
        Enumeration e = all.elements();

        while (e.hasMoreElements()) {
            NPGraphConstraints c = (NPGraphConstraints) (e.nextElement());

            c.constraintNumber = numConstraints;
            constraints[numConstraints] = c;
            numConstraints++;
        }
    }


    /**
     * This must be called <i>after</i> the NPTypes and NPFunctionSets
     * have been set up.
     */
    public final void setup(final EvolutionState state,
                            final Parameter base) {

        // What's my name?
        name = state.parameters.getString(base.push(P_NAME), null);

        if (name == null)
            state.output.fatal("No name was given for this function set.",
                               base.push(P_NAME));

            // Register me

        NPGraphConstraints old_constraints = (NPGraphConstraints) (all.put(name,
                this));

        if (old_constraints != null)
            state.output.fatal("The NP graph constraint \"" + name
                               + "\" has been defined multiple times.", base.push(P_NAME));

            // Load my initializing builder

        init =
            (NPGraphBuilder) (state.parameters.getInstanceForParameter(base.push(P_INIT),
                null, NPGraphBuilder.class));

        init.setup(state, base.push(P_INIT));

        // Load my return type
        String s = state.parameters.getString(base.push(P_VALUE), null);

        if (s == null)
            state.output.fatal("No value type given for the NPGraphConstraints "
                               + name, base.push(P_VALUE));

        graphtype = NPType.typeFor(s, state);

        // Load my function set

        s = state.parameters.getString(base.push(P_FUNCTIONSET), null);

        if (s == null)
            state.output.fatal("No function set given for the NPGraphConstraints "
                               + name, base.push(P_VALUE));

        functionset = NPFunctionSet.functionSetFor(s, state);

        state.output.exitIfErrors();// otherwise checkFunctionSetValidity might crash below

        // Determine the validity of the function set
        // the way we do that is by gathering all the types that
        // are transitively used, starting with graphtype, as in:
        Hashtable typ = new Hashtable();

        checkFunctionSetValidity(typ, graphtype);

        // next we make sure that for every one of these types,
        // there's a terminal with that return type, and *maybe*
        // a nonterminal
        Enumeration e = typ.elements();

        while (e.hasMoreElements()) {
            NPType t = (NPType) (e.nextElement());
            NPFuncInfo[] i = functionset.terminals[t.type];

            if (i.length == 0)  // uh oh
                state.output.error("In function set " + functionset
                                   + " for the NPGraphConstraints " + this
                                   + ", no terminals are given with the return type "
                                   + t
                                   + " which is required by other functions in the function set.", base);

            i = functionset.nonterminals[t.type];

            if (i.length == 0)  // uh oh
                state.output.warning("In function set " + functionset
                                     + " for the NPGraphConstraints " + this
                                     + ", no *nonterminals* are given with the return type "
                                     + t
                                     + " which is required by other functions in the function set.  This may or may not be a problem for you.", base);
        }

        state.output.exitIfErrors();
    }

    // When completed, done will hold all the types which are needed
    // in the function set -- you can then check to make sure that
    // they contain at least one terminal and (hopefully) at least
    // one nonterminal.

    private void checkFunctionSetValidity(final Hashtable done,
                                          final NPType type) {

        // Grab the array in nodes
        NPFuncInfo[] i = functionset.nodes[type.type];

        // For each argument type in a node in i, if it's not in done,
        // then add it to done and call me on it
        for (int x = 0; x < i.length; x++)
            for (int y = 0; y < i[x].node.constraints().linktypes.length;
                    y++)
                if (done.get(i[x].node.constraints().linktypes[y]) == null) {
                    done.put(i[x].node.constraints().linktypes[y],
                             i[x].node.constraints().linktypes[y]);
                    checkFunctionSetValidity(done,
                                             i[x].node.constraints().linktypes[y]);
                }
    }


    /**
     * You must guarantee that after calling constraintsFor(...) one or
     * several times, you call state.output.exitIfErrors() once.
     */

    public static NPGraphConstraints constraintsFor(final String name,
            final EvolutionState state) {
        NPGraphConstraints constraints = (NPGraphConstraints) (all.get(name));

        if (constraints == null)
            state.output.error("The NP graph constraint \"" + name
                               + "\" could not be found.");

        return constraints;
    }


    public final void writeObject(ObjectOutputStream out) throws IOException {

        // this wastes an hashtable pointer, but what the heck.

        out.defaultWriteObject();
        out.writeObject(all);
    }

    public final void readObject(ObjectInputStream in)
            throws IOException, ClassNotFoundException {
        in.defaultReadObject();

        all = (Hashtable) (in.readObject());
    }

}

