/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package evision.ec.np;

import java.util.*;
import java.io.*;

import ec.*;
import ec.util.*;

/**
 * Title:
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author
 * @version 1.0
 */

/**
 * NPIndividual is an Individual used for NP evolution runs.
 * NPIndividuals contain, at the very least, a nonempty array of NPTrees.
 * You can use NPIndividual directly, or subclass it to extend it as
 * you see fit.
 * <p><b>Parameters</b><br>
 * <table>
 * <tr><td valign=top><i>base</i>.<tt>numtrees</tt><br>
 * <font size=-1>int &gt;= 1</font></td>
 * <td valign=top>(number of graphs in the NPIndividual)</td></tr>
 * <tr><td valign=top><i>base</i>.<tt>tree.</tt><i>n</i><br>
 * <font size=-1>classname, inherits or = ec.gp.NPTree</font></td>
 * <td valign=top>(class of tree <i>n</i> in the individual)</td></tr>
 * </table>
 * <p><b>Default Base</b><br>
 * gp.individual
 * <p><b>Parameter bases</b><br>
 * <table>
 * <tr><td valign=top><i>base</i>.<tt>tree.</tt><i>n</i></td>
 * <td>tree <i>n</i> in the individual</td></tr>
 * </table>
 *
 * @author Sean Luke
 * @version 1.0
 */

public class NPIndividual extends Individual {
    static final long serialVersionUID = 8425858718585166509L;

    public static final String P_INDIVIDUAL = "individual";
    public static final String P_NUMGRAPHS = "numgraphs";
    public static final String P_GRAPH = "graph";

    public static final String EVALUATED_PREAMBLE = "Evaluated: ";

    public NPGraph[] graphs;

    public Parameter defaultBase() {
        return NPDefaults.base().push(P_INDIVIDUAL);
    }

    public boolean equals(Object ind) {
        if (!(this.getClass().equals(ind.getClass())))
            return false;   // NPIndividuals are special.

        NPIndividual i = (NPIndividual) ind;

        if (graphs.length != i.graphs.length)
            return false;

            // this default version works fine for most NPIndividuals.

        for (int x = 0; x < graphs.length; x++)
            if (!(graphs[x].graphEquals(i.graphs[x])))
                return false;

        return true;
    }

    public int hashCode() {

        int hash = this.getClass().hashCode();

        return hash;
    }

    /**
     * Sets up a prototypical NPIndividual with those features which it
     * shares with other NPIndividuals in its species, and nothing more.
     */

    public void setup(final EvolutionState state, final Parameter base) {
        Parameter def = defaultBase();

        // set my evaluation to false
        evaluated = false;

        // how many graphs?
        int t = state.parameters.getInt(base.push(P_NUMGRAPHS),
                                        def.push(P_NUMGRAPHS),
                                        1);

        if (t <= 0)
            state.output.fatal("A NPIndividual have at least one graph.",
                               base.push(P_NUMGRAPHS), def.push(P_NUMGRAPHS));

            // load the graphs

        graphs = new NPGraph[t];

        for (int x = 0; x < t; x++) {
            Parameter p = base.push(P_GRAPH).push("" + x);

            graphs[x] =
                (NPGraph) (state.parameters.getInstanceForParameterEq(p,
                    def.push(P_GRAPH).push("" + x), NPGraph.class));
            graphs[x].owner = this;

            graphs[x].setup(state, p);
        }

        // now that our function sets are all associated with graphs,
        // give the nodes a chance to determine whether or not this is
        // going to work for them (especially the ADFs).

        for (int x = 0; x < t; x++) {
            for (int w = 0;
                    w < graphs[x].constraints().functionset.nodes.length;
                    w++) {
                NPFuncInfo[] gpfi =
                    graphs[x].constraints().functionset.nodes[w];

                for (int y = 0; y < gpfi.length; y++)
                    gpfi[y].node.checkConstraints(state, x, this, base);
            }
        }

        // because I promised with checkConstraints(...)
        state.output.exitIfErrors();
    }


    // Mainly for debugging
    public void verifyIndividual(boolean unconnectedNode) {
        for (int i = 0; i < graphs.length; i++) {
            graphs[i].verifyGraph(unconnectedNode);
        }
    }

    public void printIndividualForHumans(final EvolutionState state,
                                         final int log, final int verbosity) {

        state.output.println(EVALUATED_PREAMBLE
                             + (evaluated ? "true" : "false"), verbosity,
                             log);

        fitness.printFitnessForHumans(state, log, verbosity);

        for (int i = 0; i < graphs.length; i++) {
            state.output.println("Graph " + i + ":", verbosity, log);
            graphs[i].printGraphForHumans(state, log, verbosity);
            state.output.println("", verbosity, log);
        }
    }

    public void printIndividual(final EvolutionState state, final int log,
                                final int verbosity) {
        throw new RuntimeException("It should be implemented");
    }

    public void printIndividual(final EvolutionState state, final int thread,
                                final PrintWriter writer) {
        throw new RuntimeException("It should be implemented");
    }

    public void readIndividual(final EvolutionState state, final int thread,
                               final LineNumberReader reader) throws IOException,
                               CloneNotSupportedException {
        throw new RuntimeException("It should be implemented");
    }

    public Object protoClone() throws CloneNotSupportedException {
        NPIndividual newInd = (NPIndividual) (super.protoClone());

        // deep-clone
        newInd.graphs = new NPGraph[graphs.length];

        for (int i = 0; i < graphs.length; i++) {
            newInd.graphs[i] = (NPGraph) (graphs[i].protoClone());
            newInd.graphs[i].owner = newInd;
        }

        return newInd;
    }

}

