/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package evision.ec.np;

import java.io.PrintWriter;
import java.io.Serializable;

import ec.*;
import ec.util.*;

import ec.EvolutionState;
import ec.Species;


/**
 * Title:
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author
 * @version 1.0
 */

/**
 * NPNode is a NPNodeParent which is the abstract superclass of
 * all NP function nodes in graphs.  NPNode contains quite a few functions
 * for cloning subgraphs in special ways, counting the number of nodes
 * in subgraphs in special ways, and finding specific nodes in subgraphs.
 *
 * NPNode's protoClone() method does not clone its children (it copies the
 * array, but that's it).  If you want to deep-clone a graph or subgraph, you
 * should use one of the cloneReplacing(...) methods instead.
 *
 * <p>NPNodes contain a number of important items:
 * <ul><li>A <i>constraints</i> object which defines the name of the node,
 * its arity, and its type constraints. This
 * object is shared with all NPNodes of the same function name/arity/returntype/childtypes.
 * <li>A <i>parent</i>.  This is either another NPNode, or (if this node
 * is the root) a NPGraph.
 * <li>Zero or more <i>children</i>, which are NPNodes.
 * <li>An argument position in its parent.
 * </ul>
 *
 * <p><b>Parameters</b><br>
 * <table>
 * <tr><td valign=top><i>base</i>.<tt>nc</tt><br>
 * <font size=-1>String</font></td>
 * <td valign=top>(name of the node constraints for the NPNode)</td></tr>
 * </table>
 * <p><b>Default Base</b><br>
 * gp.node
 *
 * @author Sean Luke
 * @version 1.0
 */

public abstract class NPNode implements NPNodeParent {
    static final long serialVersionUID = -8915532376213040435L;

    public static final String P_NODE = "node";
    public static final String P_NODECONSTRAINTS = "nc";
    public static final String NPNODEPRINTTAB = "    ";
    public static final int MAXPRINTBYTES = 40;

    public static final int NODESEARCH_ALL = 0;
    public static final int NODESEARCH_TERMINALS = 1;
    public static final int NODESEARCH_NONTERMINALS = 2;
    public static final int NODESEARCH_CUSTOM = 3;

    public static final int SITUATION_NEWIND = 0;
    public static final int SITUATION_MUTATION = 1;

    public static final char REPLACEMENT_CHAR = '@';

    public NPGraph owner;

    public NPNode links[];

    /**
     * The NPNode's memory.
     */
    public NPMemory memory;

    /**
     * The NPNode's position in node array list.
     */
    public int position;

    /**
     * The NPNode's constraints.
     */
    public byte constraints;

    /* Returns the NPNode's constraints.  A good JIT compiler should inline this. */
    public final NPNodeConstraints constraints() {
        return NPNodeConstraints.constraints[constraints];
    }

    /**
     * The default base for NPNodes -- defined even though
     * NPNode is abstract so you don't have to in subclasses.
     */
    public Parameter defaultBase() {
        return NPDefaults.base().push(P_NODE);
    }

    /**
     * You ought to override this method to check to make sure that the
     * constraints are valid as best you can tell.  Things you might
     * check for:
     * <ul>
     * <li> children.length is correct
     * <li> certain arguments in constraints.childtypes are
     * swap-compatible with each other
     * <li> constraints.returntype is swap-compatible with appropriate
     * arguments in constraints.childtypes
     * </ul>
     * You can't check for everything, of course, but you might try some
     * obvious checks for blunders.  The default version of this method
     * is empty for now, but you should still call super.checkConstraints(state)
     * just to be certain.
     * The ultimate caller of this method must guarantee that he will eventually
     * call state.output.exitIfErrors(), so you can freely use state.output.error
     * instead of state.output.fatal(), which will help a lot.
     * Warning: this method may get called more than once.
     */

    public void checkConstraints(final EvolutionState state, final int graph,
                                 final NPIndividual typicalIndividual,
                                 final Parameter individualBase) {}

    /**
     * Sets up a <i>prototypical</i> NPNode with those features all nodes of that
     * prototype share, and nothing more.  So no filled-in children,
     * no argposition, no parent.  Yet.
     * This must be called <i>after</i> the NPTypes and NPNodeConstraints
     * have been set up.  Presently they're set up in NPInitializer,
     * which gets called before this does, so we're safe.
     * You should override this if you need to load some special features on
     * a per-function basis.  Note that base hangs off of a function set, so
     * this method may get called for different instances in the same NPNode
     * class if they're being set up as prototypes for different NPFunctionSets.
     * If you absolutely need some global base, then you should use something
     * hanging off of NPDefaults.base().
     * The ultimate caller of this method must guarantee that he will eventually
     * call state.output.exitIfErrors(), so you can freely use state.output.error
     * instead of state.output.fatal(), which will help a lot.
     */

    public void setup(final EvolutionState state, final Parameter base) {
        Parameter def = defaultBase();

        // determine my constraints -- at this point, the constraints should have been loaded.
        String s = state.parameters.getString(base.push(P_NODECONSTRAINTS),
                                              def.push(P_NODECONSTRAINTS));

        if (s == null)
            state.output.fatal("No node constraints are defined for the NPNode "
                               + toStringForError(),
                                                     base.push(P_NODECONSTRAINTS),
                                                     def.push(P_NODECONSTRAINTS));
        else
            constraints = NPNodeConstraints.constraintsFor(s,
                    state).constraintNumber;

        // The number of children is determined by the constraints.  Though
        // for some special versions of NPNode, we may have to enforce certain
        // rules, checked in children versions of setup(...)

        links = new NPNode[constraints().linktypes.length];
    }


    /**
     * Returns true if I can swap into node's position.
     */

    public final boolean swapCompatibleWith(final NPNode node) {

        // I'm atomically compatible with him; a fast check
        if (constraints().valuetype
                == node.constraints().valuetype)   // no need to check for compatibility
            return true;
        else
            return false;
    }


    public void resetMemory(final EvolutionState state, final int thread) {

        // NPNodes may well be memoryless
        if (memory != null) {
            memory.resetMemory(state, thread);
        }
    }

    /**
     * Starts a node in a new life immediately after it has been cloned.
     * The default version of this function does nothing.  The purpose of
     * this function is to give ERCs a chance to set themselves to a new
     * random value after they've been cloned from the prototype.
     * You should not assume that the node is properly connected to other
     * nodes in the graph at the point this method is called.
     */

    public void resetNode(final EvolutionState state, final int thread) {}

    /**
     * A convenience function for identifying a NPNode in an error message
     */
    public String errorInfo() {
        return "NPNode " + toString() + " in the function set for graph ";
    }


    public Object protoClone() throws CloneNotSupportedException {
        NPNode obj = (NPNode) (super.clone());

        obj.links = new NPNode[links.length];

        // NPNodes may well be memoryless
        if (memory != null) {
            obj.memory = (NPMemory) memory.protoClone();
        }

        return obj;
    }

    public final Object protoCloneSimple() {
        try {
            return protoClone();
        }
        catch (CloneNotSupportedException e) {
            throw new InternalError();
        }   // never happens
     }


    /**
     * Returns a hashcode usually associated with all nodes that are
     * equal to you (using nodeEquals(...)).  The default form
     * of this method returns the hashcode of the node's class.
     * ERCs in particular probably will want to override this method.
     */
    public int nodeHashCode() {
        return (this.getClass().hashCode());
    }

    /**
     * Returns true if I am the "genetically" same as this node, and our
     * children arrays are the same length, though
     * we may have different parents and children.  The default form
     * of this method does a class comparison.  You may need
     * to override this to perform special comparisons, if you're
     * an ERC, ADF, or ADM for example.
     */
    public boolean nodeEquals(final NPNode node) {
        return (this.getClass().equals(node.getClass())
                && links.length == node.links.length);
    }

    /**
     * Prints out a human-readable and Lisp-like atom for the node,
     * and returns the number of bytes in the string that you sent
     * to the log (use print(),
     * not println()).  The default version gets the atom from
     * toStringForHumans().
     */
    public int printNodeForHumans(final EvolutionState state, final int log,
                                  final int verbosity) {
        String n = toStringForHumans();

        state.output.print(n, verbosity, log);

        return n.length();
    }


    /**
     * Prints out a COMPUTER-readable and Lisp-like atom for the node, which
     * is also suitable for readNode to read, and returns
     * the number of bytes in the string that you sent to the log (use print(),
     * not println()).  The default version gets the atom from toString().
     * O(1).
     */
    public int printNode(final EvolutionState state, final int log,
                         final int verbosity) {
        String n = toString();

        state.output.print(n, verbosity, log);

        return n.length();
    }


    /**
     * Prints out a COMPUTER-readable and Lisp-like atom for the node, which
     * is also suitable for readNode to read, and returns
     * the number of bytes in the string that you sent to the log (use print(),
     * not println()).  The default version gets the atom from toString().
     * O(1).
     */

    public int printNode(final EvolutionState state, final int thread,
                         final PrintWriter writer) {
        String n = toString();

        writer.print(n);

        return n.length();
    }


    /**
     * Returns a Lisp-like atom for the node which can be read in again by computer.
     * If you need to encode an integer or a float or whatever for some reason
     * (perhaps if it's an ERC), you should use the ec.util.Code library.
     */

    public abstract String toString();

    /**
     * Returns a Lisp-like atom for the node which is intended for human
     * consumption, and not to be read in again.  The default version
     * just calls toString().
     */

    public String toStringForHumans() {
        return toString();
    }

    /**
     * Returns a description of the node that can make it easy to identify
     * in error messages (by default, at least its name and the graph it's found in).
     * It's okay if this is a reasonably expensive procedure -- it won't be called
     * a lot.
     */

    public String toStringForError() {
        return toString();
    }



    /**
     * Evaluates the node with the given thread, state, individual, problem, and stack.
     * Your random number generator will be state.random[thread].
     */

    public abstract void eval(final EvolutionState state, final int thread,
                              final NPData input, final ADFContext context,
                              final NPMemory newMemory,
                              final NPIndividual individual,
                              final Problem problem);
}

