/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package evision.ec.np;

import java.util.*;
import java.io.*;

import ec.*;
import ec.util.*;

/**
 * Title:
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author
 * @version 1.0
 */


/**
 * A NPNodeConstraints is a Clique which defines constraint information
 * common to many different NPNode functions, namely return types,
 * child types, and number of children.
 * NPNodeConstraints have unique names by which
 * they are identified.
 *
 * <p>In adding new things to NPNodeConstraints, you should ask yourself
 * the following questions: first, is this something that takes up too
 * much memory to store in NPNodes themselves?  second, is this something
 * that needs to be accessed very rapidly, so cannot be implemented
 * as a method call in a NPNode?  third, can this be shared among
 * different NPNodes, even ones representing different functions?
 * <p><b>Parameters</b><br>
 * <table>
 * <tr><td valign=top><i>base</i>.<tt>size</tt><br>
 * <font size=-1>int &gt;= 1</font></td>
 * <td valign=top>(number of node constraints)</td></tr>
 * <tr><td valign=top><i>base.n</i>.<tt>name</tt><br>
 * <font size=-1>String</font></td>
 * <td valign=top>(name of node constraint <i>n</i>)</td></tr>
 * <tr><td valign=top><i>base.n</i>.<tt>returns</tt><br>
 * <font size=-1>String</font></td>
 * <td valign=top>(return type for node constraint <i>n</i>)</td></tr>
 * <tr><td valign=top><i>base.n</i>.<tt>size</tt><br>
 * <font size=-1>int &gt;= 1</font></td>
 * <td valign=top>(number of child arguments for node constraint <i>n</i>)</td></tr>
 * <tr><td valign=top><i>base.n</i>.<tt>child.</tt><i>m</i><br>
 * <font size=-1>String</font></td>
 * <td valign=top>(name of type for child argument <i>m</i> of node constraint <i>n</i>)</td></tr>
 * <tr><td valign=top><i>base</i>.<tt>prob</tt><br>
 * <font size=-1>float &gt;= 0.0</font></td>
 * <td valign=top>(auxillary probability of selection -- used by ec.gp.build.PTC1 and ec.gp.build.PTC2)</td></tr>
 * </table>
 * @author Sean Luke
 * @version 1.0
 */

public class NPNodeConstraints implements Clique {
    static final long serialVersionUID = -2958335456080750873L;

    public static final int SIZE_OF_BYTE = 256;
    public final static String P_NAME = "name";
    public final static String P_VALUE = "value";
    public final static String P_LINK = "link";
    public final static String P_SIZE = "size";
    public final static String P_PROBABILITY = "prob";
    public final static float DEFAULT_PROBABILITY = 1.0f;

    /**
     * Probability of selection -- an auxillary measure mostly used by PTC1/PTC2
     * right now
     */
    public float probabilityOfSelection;

    /**
     * The byte value of the constraints -- we can only have 256 of them
     */
    public byte constraintNumber;

    /**
     * The value type for a NPNode
     */
    public NPType valuetype;

    /**
     * The link types for a NPNode
     */
    public NPType[] linktypes;

    /**
     * The name of the NPNodeConstraints object -- this is NOT the
     * name of the NPNode
     */
    public String name;

    /**
     * A repository of all the NPNodeConstraints in the system.
     */
    public static Hashtable all;

    public static NPNodeConstraints[] constraints;
    public static byte numConstraints;

    static {
        all = new Hashtable();
        constraints = new NPNodeConstraints[SIZE_OF_BYTE];
        numConstraints = 0;
    }

    public String toString() {
        return name;
    }


    /**
     * Sets up all the NPNodeConstraints, loading them from the parameter
     * file.  This must be called before anything is called which refers
     * to a type by name.
     */

    public static void setupConstraints(final EvolutionState state,
                                        final Parameter base) {
        state.output.message("Processing NP Node Constraints");

        // How many NPNodeConstraints do we have?
        int x = state.parameters.getInt(base.push(P_SIZE), null, 1);

        if (x <= 0)
            state.output.fatal("The number of NP node constraints must be at least 1.",
                               base.push(P_SIZE));

            // Load our constraints

        for (int y = 0; y < x; y++) {
            NPNodeConstraints c;

            // Figure the constraint class
            if (state.parameters.exists(base.push("" + y)))
                c = (NPNodeConstraints) (state.parameters.getInstanceForParameterEq(base.push(""
                        + y), null, NPNodeConstraints.class));
            else {
                state.output.message("No NP Node Constraints specified, assuming the default class: evision.ec.np.NPNodeConstraints for "
                                     + base.push("" + y));

                c = new NPNodeConstraints();
            }

            c.setup(state, base.push("" + y));
        }

        // set our constraints array up
        Enumeration e = all.elements();

        while (e.hasMoreElements()) {
            NPNodeConstraints c = (NPNodeConstraints) (e.nextElement());

            c.constraintNumber = numConstraints;
            constraints[numConstraints] = c;
            numConstraints++;
        }
    }


    /**
     * This must be called <i>after</i> the NPTypes have been set up.
     */
    public final void setup(final EvolutionState state,
                            final Parameter base) {

        // What's my name?
        name = state.parameters.getString(base.push(P_NAME), null);

        if (name == null)
            state.output.fatal("No name was given for this node constraints.",
                               base.push(P_NAME));

            // Register me

        NPNodeConstraints old_constraints = (NPNodeConstraints) (all.put(name,
                this));

        if (old_constraints != null)
            state.output.fatal("The NP node constraint \"" + name
                               + "\" has been defined multiple times.", base.push(P_NAME));

            // What's my return type?

        String s = state.parameters.getString(base.push(P_VALUE), null);

        if (s == null)
            state.output.fatal("No value type given for the NPNodeConstraints "
                               + name, base.push(P_VALUE));

        valuetype = NPType.typeFor(s, state);

        // Load probability of selection

        if (state.parameters.exists(base.push(P_PROBABILITY), null)) {
            float f = state.parameters.getFloat(base.push(P_PROBABILITY),
                                                null, 0);

            if (f < 0)
                state.output.fatal("The probability of selection is < 0, which is not valid.",
                                   base.push(P_PROBABILITY), null);

            probabilityOfSelection = f;
        }
        else
            probabilityOfSelection = DEFAULT_PROBABILITY;

        // How many child types do I have?

        int x = state.parameters.getInt(base.push(P_SIZE), null, 0);

        if (x < 0)
            state.output.fatal("The number of children types for the NPNodeConstraints "
                               + name + " must be >= 0.", base.push(P_SIZE));

        linktypes = new NPType[x];

        Parameter p = base.push(P_LINK);

        // Load my children
        for (x = 0; x < linktypes.length; x++) {
            s = state.parameters.getString(p.push("" + x), null);

            if (s == null)
                state.output.fatal("Type #" + x
                                   + " is not defined for the NPNodeConstraints "
                                   + name + ".", base.push("" + x));

            linktypes[x] = NPType.typeFor(s, state);
        }

        // ...because I promised when I called typeFor(...)
        state.output.exitIfErrors();
    }


    /**
     * You must guarantee that after calling constraintsFor(...) one or
     * several times, you call state.output.exitIfErrors() once.
     */

    public static NPNodeConstraints constraintsFor(final String name,
            final EvolutionState state) {
        NPNodeConstraints constraints = (NPNodeConstraints) (all.get(name));

        if (constraints == null)
            state.output.error("The NP node constraint \"" + name
                               + "\" could not be found.");

        return constraints;
    }


    public final void writeObject(ObjectOutputStream out) throws IOException {

        // this wastes an hashtable pointer, but what the heck.

        out.defaultWriteObject();
        out.writeObject(all);
    }

    public final void readObject(ObjectInputStream in)
            throws IOException, ClassNotFoundException {
        in.defaultReadObject();

        all = (Hashtable) (in.readObject());
    }

}

